# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import torch
from lightning.pytorch.strategies import DDPStrategy


class SimpleDDPStrategy(DDPStrategy):
    """Effectively Lightning's :class:`~lightning.pytorch.strategies.DDPStrategy`, but doing manual gradient syncs instead of using PyTorch's :class:`~torch.nn.parallel.DistributedDataParallel` wrapper.

    .. note::
        To use train-time compilation with multi-rank training, this strategy must be used in place of PyTorch Lightning's :class:`~lightning.pytorch.strategies.DDPStrategy`.

    Example use in the config file:

    .. code-block:: yaml

      trainer:
        _target_: lightning.Trainer
        # other trainer arguments
        strategy:
          _target_: nequip.train.SimpleDDPStrategy
    """

    def configure_ddp(self) -> None:
        pass

    def post_backward(self, closure_loss: torch.Tensor) -> None:
        """
        Manual syncing of gradients after the backwards pass.
        """
        # cat all gradients into a single tensor for efficiency
        grad_tensors = []
        for param in self.model.parameters():
            if param.requires_grad and param.grad is not None:
                grad_tensors.append(param.grad.data.view(-1))

        if grad_tensors:
            # cat and reduce
            flat_grads = torch.cat(grad_tensors)
            # NOTE: averaging (i.e. summing and dividing by number of ranks) is consistent with PyTorch Lightning's `DDPStrategy`
            # in the training loop, we account for this by multiplying the loss by the number of ranks before the backwards call
            if torch.distributed.get_backend() == "gloo":
                torch.distributed.all_reduce(
                    flat_grads, op=torch.distributed.ReduceOp.SUM
                )
                flat_grads /= torch.distributed.get_world_size()
            else:
                torch.distributed.all_reduce(
                    flat_grads, op=torch.distributed.ReduceOp.AVG
                )

            # copy reduced gradients back
            offset = 0
            for param in self.model.parameters():
                if param.requires_grad and param.grad is not None:
                    numel = param.grad.numel()
                    param.grad.data.copy_(
                        flat_grads[offset : offset + numel].view_as(param.grad.data)
                    )
                    offset += numel



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.

"""Custom OmegaConf resolvers for nequip."""

from omegaconf import OmegaConf
from typing import Dict, Callable, Set


def _sanitize_int(x, client: str):
    err_msg = f"`{client} resolver accepts nonnegative integer inputs, but found {x}"
    if isinstance(x, str):
        assert x.isdigit(), err_msg
        x = int(x)
    assert isinstance(x, int), err_msg
    return x


def int_div(a, b):
    """Integer division resolver for OmegaConf."""
    a = _sanitize_int(a, "int_div")
    b = _sanitize_int(b, "int_div")

    if a % b != 0:
        raise ValueError(
            f"`int_div` requires exact division, but {a} is not divisible by {b}"
        )

    return a // b


def int_mul(a, b):
    """Integer multiplication resolver for OmegaConf."""
    a = _sanitize_int(a, "int_mul")
    b = _sanitize_int(b, "int_mul")
    return a * b


# === Resolver Registry ===

_DEFAULT_RESOLVERS: Dict[str, Callable] = {
    "int_div": int_div,
    "int_mul": int_mul,
}

_REGISTERED_RESOLVERS: Set[str] = set()
_DEFAULT_RESOLVERS_REGISTERED: bool = False


def register_resolvers(resolvers: Dict[str, Callable]) -> None:
    """Register custom OmegaConf resolvers.

    Args:
        resolvers (Dict[str, Callable]): mapping from resolver name to resolver function

    Raises:
        ValueError: if any resolver name conflicts with existing registered resolvers
    """
    # Check for conflicts with already registered resolvers
    new_names = set(resolvers.keys())
    conflicts = new_names.intersection(_REGISTERED_RESOLVERS)
    if conflicts:
        raise ValueError(
            f"Resolver name(s) {conflicts} already registered. Cannot register the same resolver name twice."
        )

    # Register each resolver with OmegaConf
    for name, func in resolvers.items():
        OmegaConf.register_new_resolver(name, func)
        _REGISTERED_RESOLVERS.add(name)


def _register_default_resolvers():
    """Register all default nequip OmegaConf resolvers."""
    global _DEFAULT_RESOLVERS_REGISTERED
    if not _DEFAULT_RESOLVERS_REGISTERED:
        register_resolvers(_DEFAULT_RESOLVERS)
        _DEFAULT_RESOLVERS_REGISTERED = True



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from ._base_datamodule import NequIPDataModule
from .sgdml_datamodule import sGDML_CCSD_DataModule
from ._3bpa_datamodule import NequIP3BPADataModule
from ._ase_datamodule import ASEDataModule
from .tm23_datamodule import TM23DataModule
from .rmd17_datamodule import rMD17DataModule
from .md22_datamodule import MD22DataModule
from .samd23_datamodule import SAMD23DataModule

__all__ = [
    NequIPDataModule,
    sGDML_CCSD_DataModule,
    rMD17DataModule,
    MD22DataModule,
    NequIP3BPADataModule,
    TM23DataModule,
    ASEDataModule,
    SAMD23DataModule,
]



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import torch
from nequip.data import AtomicDataDict, compute_neighborlist_
from typing import Optional, Dict, Union, List


class NeighborListTransform:
    """Constructs a neighborlist and adds it to the ``AtomicDataDict``.

    Args:
        r_max (float): cutoff radius used for nearest neighbors
        per_edge_type_cutoff (Dict): optional per-edge-type cutoffs (must be <= r_max)
        type_names (List[str]): list of atom type names
    """

    def __init__(
        self,
        r_max: float,
        per_edge_type_cutoff: Optional[
            Dict[str, Union[float, Dict[str, float]]]
        ] = None,
        type_names: Optional[List[str]] = None,
        **kwargs,
    ):
        self.r_max = r_max
        self.type_names = type_names
        self.per_edge_type_cutoff = per_edge_type_cutoff
        self.kwargs = kwargs

        # set up normalizer for per-edge-type cutoffs if provided
        self._per_edge_type = False
        if per_edge_type_cutoff is not None:
            assert (
                type_names is not None
            ), "`type_names` required for `per_edge_type_cutoff`"
            self._per_edge_type = True
            from nequip.nn.embedding import EdgeLengthNormalizer

            self._normalizer = EdgeLengthNormalizer(
                r_max=self.r_max,
                type_names=type_names,
                per_edge_type_cutoff=per_edge_type_cutoff,
            )

    def __call__(self, data: AtomicDataDict.Type) -> AtomicDataDict.Type:
        data = compute_neighborlist_(data, self.r_max, **self.kwargs)

        # prune based on per-edge-type cutoffs if specified
        if self._per_edge_type:
            if AtomicDataDict.ATOM_TYPE_KEY not in data:
                raise KeyError(
                    f"Per-edge-type cutoffs require '{AtomicDataDict.ATOM_TYPE_KEY}' to be present in the data. "
                    "This is likely because nequip.data.transforms.ChemicalSpeciesToAtomTypeMapper needs to be applied before "
                    "nequip.data.transforms.NeighborListTransform. Please check your data transform order in the configuration."
                )
            data = self._apply_per_edge_type_cutoffs(data)

        return data

    def _apply_per_edge_type_cutoffs(
        self, data: AtomicDataDict.Type
    ) -> AtomicDataDict.Type:
        """Prune neighbor list based on per-edge-type cutoffs using ``EdgeLengthNormalizer``."""

        # get mask for pruning
        # make a shallow copy of original data dict, such that the original doens't have the tensors added by applying `self._normalizer`
        mask = (
            self._normalizer(data.copy())[AtomicDataDict.NORM_LENGTH_KEY].view(-1)
            <= 1.0
        )

        # mask neighborlist
        data[AtomicDataDict.EDGE_INDEX_KEY] = data[AtomicDataDict.EDGE_INDEX_KEY][
            :, mask
        ]
        if AtomicDataDict.EDGE_CELL_SHIFT_KEY in data:
            data[AtomicDataDict.EDGE_CELL_SHIFT_KEY] = data[
                AtomicDataDict.EDGE_CELL_SHIFT_KEY
            ][mask]

        return data


class SortedNeighborListTransform(NeighborListTransform):
    """Behaves like :class:`NeighborListTransform` but additionally sorts the neighborlist and provides transpose permutation indices."""

    def __call__(self, data: AtomicDataDict.Type) -> AtomicDataDict.Type:
        # first compute the basic neighborlist
        data = super().__call__(data)

        # sort the edge index and corresponding edge attributes
        edge_idxs = data[AtomicDataDict.EDGE_INDEX_KEY]

        # short-circuit for empty edge index case
        if edge_idxs.numel() == 0:
            return data

        sort_indices = torch.argsort(
            edge_idxs[0] * AtomicDataDict.num_nodes(data) + edge_idxs[1]
        )

        # sort edge indices
        data[AtomicDataDict.EDGE_INDEX_KEY] = torch.index_select(
            edge_idxs, 1, sort_indices
        )
        edge_idxs = data[AtomicDataDict.EDGE_INDEX_KEY]

        # sort cell shifts if present
        if AtomicDataDict.EDGE_CELL_SHIFT_KEY in data:
            data[AtomicDataDict.EDGE_CELL_SHIFT_KEY] = torch.index_select(
                data[AtomicDataDict.EDGE_CELL_SHIFT_KEY], 0, sort_indices
            )

        # compute transpose permutation for backward pass
        # compute full transpose permutation: sort by sender first, then receiver
        transpose_perm = torch.argsort(
            edge_idxs[1] * AtomicDataDict.num_nodes(data) + edge_idxs[0]
        )
        data[AtomicDataDict.EDGE_TRANSPOSE_PERM_KEY] = transpose_perm

        return data



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.

import copy
import tempfile
import pathlib
import subprocess
import os
import sys

from omegaconf import OmegaConf, open_dict


def _check_and_print(retcode):
    """Helper function to check subprocess return code and print output on failure."""
    __tracebackhide__ = True
    if retcode.returncode:
        if retcode.stdout is not None and len(retcode.stdout) > 0:
            print(retcode.stdout.decode("ascii", errors="replace"))
        if retcode.stderr is not None and len(retcode.stderr) > 0:
            print(retcode.stderr.decode("ascii", errors="replace"), file=sys.stderr)
        retcode.check_returncode()


def _training_session(
    conffile,
    model_dtype,
    extra_train_from_save=None,
    model_config=None,
    training_module=None,
):
    """
    Create a training session using config files with optional model injection.

    Unified function for both integration and unit test training sessions.

    Args:
        conffile: Name of config file (e.g., "minimal_aspirin.yaml")
        model_dtype: Model dtype string (e.g., "float32", "float64")
        extra_train_from_save: Optional, None/"checkpoint"/"package" for additional training
        model_config: Optional model config dict to inject (for unit tests)
        training_module: Optional training module target (for integration tests)

    Yields:
        tuple: (config, tmpdir, env) - training config, temp directory, and env vars
    """
    # find the config file in the same directory as this utils file
    current_file = pathlib.Path(__file__)
    config_path = current_file.parent / conffile
    assert (
        config_path.exists()
    ), f"Could not find config file: {conffile} in path: {config_path}"

    config = OmegaConf.load(config_path)

    with tempfile.TemporaryDirectory() as tmpdir:
        with tempfile.TemporaryDirectory() as data_tmpdir:
            if "data_source_dir" in config.data:
                config.data.data_source_dir = data_tmpdir

            # configure training module and model
            if training_module is not None:
                # integration test case: use provided training module
                config.training_module._target_ = training_module
                config.training_module.model.model_dtype = model_dtype
            else:
                # unit test case: default training module
                config.training_module._target_ = "nequip.train.EMALightningModule"

            if model_config is not None:
                # unit test case: replace model entirely
                config.training_module.model = copy.deepcopy(model_config)
                config.training_module.model.model_dtype = model_dtype

            # set training parameters
            config.data.val_dataloader.batch_size = 1
            if model_config is not None:
                # unit tests: minimal training
                config.trainer.max_epochs = 2

            with open_dict(config):
                config["hydra"] = {"run": {"dir": tmpdir}}
                # mitigate nondeterminism for tests
                config.trainer["num_sanity_val_steps"] = 0
                config.trainer["deterministic"] = True

            config = OmegaConf.create(config)
            config_path = tmpdir + "/conf.yaml"
            OmegaConf.save(config=config, f=config_path)

            # == train model ==
            env = dict(os.environ)
            env.update({"TEST_VAL_TWO": "2"})
            retcode = subprocess.run(
                ["nequip-train", "-cn", "conf"],
                cwd=tmpdir,
                env=env,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            _check_and_print(retcode)

            yield config, tmpdir, env

            # handle extra training from save (for integration tests)
            if extra_train_from_save is not None:
                with tempfile.TemporaryDirectory() as new_tmpdir:
                    new_config = config.copy()
                    with open_dict(new_config):
                        new_config["hydra"] = {"run": {"dir": new_tmpdir}}
                    if extra_train_from_save == "checkpoint":
                        with open_dict(new_config):
                            new_config["training_module"]["model"] = {
                                "_target_": "nequip.model.ModelFromCheckpoint",
                                "checkpoint_path": f"{tmpdir}/last.ckpt",
                            }
                    elif extra_train_from_save == "package":
                        # package model
                        package_path = f"{new_tmpdir}/orig_package_model.nequip.zip"
                        retcode = subprocess.run(
                            [
                                "nequip-package",
                                "build",
                                "--ckpt-path",
                                f"{tmpdir}/last.ckpt",
                                "--output-path",
                                package_path,
                            ],
                            cwd=new_tmpdir,
                            env=env,
                        )
                        _check_and_print(retcode)
                        assert pathlib.Path(
                            package_path
                        ).is_file(), "`nequip-package` didn't create file"
                        # update config
                        with open_dict(new_config):
                            new_config["training_module"]["model"] = {
                                "_target_": "nequip.model.ModelFromPackage",
                                "package_path": package_path,
                            }
                    new_config = OmegaConf.create(new_config)
                    config_path = f"{new_tmpdir}/conf.yaml"
                    OmegaConf.save(config=new_config, f=config_path)
                    retcode = subprocess.run(
                        ["nequip-train", "-cn", "conf"],
                        cwd=new_tmpdir,
                        env=env,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                    )
                    _check_and_print(retcode)
                    yield new_config, new_tmpdir, env



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from nequip.nn import SequentialGraphNetwork, AtomwiseReduce, ForceStressOutput
from nequip.nn.embedding import (
    EdgeLengthNormalizer,
    AddRadialCutoffToData,
    PolynomialCutoff,
)
from nequip.data import AtomicDataDict
from nequip.nn.pair_potential import ZBL
from .utils import model_builder

from typing import Optional, Dict, Union, Sequence


@model_builder
def ZBLPairPotential(
    r_max: float,
    type_names: Sequence[str],
    chemical_species: Sequence[str],
    units: str,
    polynomial_cutoff_p: int = 6,
    per_edge_type_cutoff: Optional[Dict[str, Union[float, Dict[str, float]]]] = None,
):
    """
    Model builder for a force field containing only a ZBL pair potential term, mainly for internal testing purposes.
    """
    edge_norm = EdgeLengthNormalizer(
        r_max=r_max,
        type_names=type_names,
        per_edge_type_cutoff=per_edge_type_cutoff,
    )
    cutoff = AddRadialCutoffToData(
        cutoff=PolynomialCutoff(polynomial_cutoff_p),
        irreps_in=edge_norm.irreps_out,
    )
    zbl_module = ZBL(
        type_names=type_names,
        chemical_species=chemical_species,
        units=units,
        irreps_in=cutoff.irreps_out,
    )
    energy_sum = AtomwiseReduce(
        reduce="sum",
        field=AtomicDataDict.PER_ATOM_ENERGY_KEY,
        out_field=AtomicDataDict.TOTAL_ENERGY_KEY,
        irreps_in=zbl_module.irreps_out,
    )
    energy_model = SequentialGraphNetwork(
        {
            "edge_norm": edge_norm,
            "cutoff": cutoff,
            "pair_potential": zbl_module,
            "total_energy_sum": energy_sum,
        }
    )
    model = ForceStressOutput(func=energy_model)

    return model



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from typing import Optional, Iterator

import torch
from torch.utils.data import Sampler


class PartialSampler(Sampler[int]):
    r"""Samples elements without replacement, but divided across a number of calls to `__iter__`.

    To ensure deterministic reproducibility and restartability, dataset permutations are generated
    from a combination of the overall seed and the epoch number. As a result, the caller must
    tell this sampler the epoch number before each time `__iter__` is called by calling
    `my_partial_sampler.step_epoch(epoch_number_about_to_run)` each time.

    This sampler decouples epochs from the dataset size and cycles through the dataset over as
    many (partial) epochs as it may take. As a result, the _dataset_ epoch can change partway
    through a training epoch.

    Args:
        data_source (Dataset): dataset to sample from
        shuffle (bool): whether to shuffle the dataset each time the _entire_ dataset is consumed
        num_samples_per_epoch (int): number of samples to draw in each call to `__iter__`.
            If `None`, defaults to `len(data_source)`.
        generator (Generator): Generator used in sampling.
    """

    data_source: torch.utils.data.Dataset
    num_samples_per_epoch: int
    shuffle: bool
    _epoch: int
    _prev_epoch: int

    def __init__(
        self,
        data_source: torch.utils.data.Dataset,
        shuffle: bool = True,
        num_samples_per_epoch: Optional[int] = None,
        generator=None,
    ) -> None:
        self.data_source = data_source
        self.shuffle = shuffle
        if num_samples_per_epoch is None:
            num_samples_per_epoch = self.num_samples_total
        self.num_samples_per_epoch = num_samples_per_epoch
        assert self.num_samples_per_epoch <= self.num_samples_total
        assert self.num_samples_per_epoch >= 1
        self.generator = generator
        self._epoch = None
        self._prev_epoch = None

    @property
    def num_samples_total(self) -> int:
        # dataset size might change at runtime
        return len(self.data_source)

    def step_epoch(self, epoch: int) -> None:
        self._epoch = epoch

    def __iter__(self) -> Iterator[int]:
        assert self._epoch is not None
        assert (self._prev_epoch is None) or (self._epoch == self._prev_epoch + 1)
        assert self._epoch >= 0

        full_epoch_i, start_sample_i = divmod(
            # how much data we've already consumed:
            self._epoch * self.num_samples_per_epoch,
            # how much data there is the dataset:
            self.num_samples_total,
        )

        if self.shuffle:
            temp_rng = torch.Generator()
            # Get new randomness for each _full_ time through the dataset
            # This is deterministic w.r.t. the combination of dataset seed and epoch number
            # Both of which persist across restarts
            # (initial_seed() is restored by set_state())
            temp_rng.manual_seed(self.generator.initial_seed() + full_epoch_i)
            full_order_this = torch.randperm(self.num_samples_total, generator=temp_rng)
            # reseed the generator for the _next_ epoch to get the shuffled order of the
            # _next_ dataset epoch to pad out this one for completing any partial batches
            # at the end:
            temp_rng.manual_seed(self.generator.initial_seed() + full_epoch_i + 1)
            full_order_next = torch.randperm(self.num_samples_total, generator=temp_rng)
            del temp_rng
        else:
            full_order_this = torch.arange(self.num_samples_total)
            # without shuffling, the next epoch has the same sampling order as this one:
            full_order_next = full_order_this

        full_order = torch.cat((full_order_this, full_order_next), dim=0)
        del full_order_next, full_order_this

        this_segment_indexes = full_order[
            start_sample_i : start_sample_i + self.num_samples_per_epoch
        ]
        # because we cycle into indexes from the next dataset epoch,
        # we should _always_ be able to get num_samples_per_epoch
        assert len(this_segment_indexes) == self.num_samples_per_epoch
        yield from this_segment_indexes

        self._prev_epoch = self._epoch

    def __len__(self) -> int:
        return self.num_samples_per_epoch



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import ase
import torch
from e3nn import o3
from e3nn.util.test import equivariance_error

from nequip.nn import GraphModuleMixin, GraphModel
from nequip.data import (
    from_dict,
    from_ase,
    compute_neighborlist_,
    AtomicDataDict,
    _NODE_FIELDS,
    _EDGE_FIELDS,
    _CARTESIAN_TENSOR_FIELDS,
)
from .dtype import dtype_from_name

from functools import wraps
from typing import Union, Optional, List


# The default float tolerance
FLOAT_TOLERANCE = {
    t: torch.as_tensor(v, dtype=dtype_from_name(t))
    for t, v in {"float32": 1e-3, "float64": 1e-10}.items()
}
# Allow lookup by name or dtype object:
for t, v in list(FLOAT_TOLERANCE.items()):
    FLOAT_TOLERANCE[dtype_from_name(t)] = v
del t, v

# This has to be somewhat large because of float32 sum reductions over many edges/atoms
PERMUTATION_FLOAT_TOLERANCE = {torch.float32: 1e-4, torch.float64: 1e-10}


# https://discuss.pytorch.org/t/how-to-quickly-inverse-a-permutation-by-using-pytorch/116205/4
def _inverse_permutation(perm):
    inv = torch.empty_like(perm)
    inv[perm] = torch.arange(perm.size(0), device=perm.device)
    return inv


def assert_permutation_equivariant(
    func: GraphModuleMixin,
    data_in: AtomicDataDict.Type,
    tolerance: Optional[float] = None,
    raise_error: bool = True,
) -> str:
    r"""Test the permutation equivariance of ``func``.

    Standard fields are assumed to be equivariant to node or edge permutations according to their standard interpretions; all other fields are assumed to be invariant to all permutations. Non-standard fields can be registered as node/edge permutation equivariant using ``register_fields``.

    Raises ``AssertionError`` if issues are found.

    Args:
        func: the module or model to test
        data_in: the example input data to test with
    """
    # Prevent pytest from showing this function in the traceback
    __tracebackhide__ = True

    if tolerance is None:
        atol = PERMUTATION_FLOAT_TOLERANCE[
            (
                func.model_dtype
                if isinstance(func, GraphModel)
                else torch.get_default_dtype()
            )
        ]
    else:
        atol = tolerance

    data_in = data_in.copy()
    device = data_in[AtomicDataDict.POSITIONS_KEY].device

    node_permute_fields = _NODE_FIELDS
    edge_permute_fields = _EDGE_FIELDS

    # Make permutations and make sure they are not identities
    n_node: int = len(data_in[AtomicDataDict.POSITIONS_KEY])
    while True:
        node_perm = torch.randperm(n_node, device=device)
        if n_node <= 1:
            break  # otherwise inf loop
        if not torch.all(node_perm == torch.arange(n_node, device=device)):
            break
    n_edge: int = data_in[AtomicDataDict.EDGE_INDEX_KEY].shape[1]
    while True:
        edge_perm = torch.randperm(n_edge, device=device)
        if n_edge <= 1:
            break  # otherwise inf loop
        if not torch.all(edge_perm == torch.arange(n_edge, device=device)):
            break
    # ^ note that these permutations are maps from the "to" index to the "from" index
    # because we index by them, the 0th element of the permuted array will be the ith
    # of the original array, where i = perm[0]. Thus i is "from" and 0 is to, so perm
    # interpreted as a map is a map from "to" to "from".

    perm_data_in = {}
    for k in data_in.keys():
        if k in node_permute_fields:
            perm_data_in[k] = data_in[k][node_perm]
        elif k in edge_permute_fields:
            perm_data_in[k] = data_in[k][edge_perm]
        else:
            perm_data_in[k] = data_in[k]

    perm_data_in[AtomicDataDict.EDGE_INDEX_KEY] = _inverse_permutation(node_perm)[
        data_in[AtomicDataDict.EDGE_INDEX_KEY]
    ][:, edge_perm]

    out_orig = func(data_in)
    out_perm = func(perm_data_in)

    assert set(out_orig.keys()) == set(
        out_perm.keys()
    ), "Permutation changed the set of fields returned by model"

    messages = []
    num_problems: int = 0
    for k in out_orig.keys():
        if k in node_permute_fields:
            err = (out_orig[k][node_perm] - out_perm[k]).abs().max()
            fail = not torch.allclose(out_orig[k][node_perm], out_perm[k], atol=atol)
            if fail:
                num_problems += 1
            messages.append(
                f"   node permutation equivariance of field {k:22}       -> max error={err:.3e}{'  FAIL' if fail else ''}"
            )
        elif k in edge_permute_fields:
            err = (out_orig[k][edge_perm] - out_perm[k]).abs().max()
            fail = not torch.allclose(out_orig[k][edge_perm], out_perm[k], atol=atol)
            if fail:
                num_problems += 1
            messages.append(
                f"   edge permutation equivariance of field {k:22}       -> max error={err:.3e}{'  FAIL' if fail else ''}"
            )
        elif k in [AtomicDataDict.EDGE_INDEX_KEY, AtomicDataDict.EDGE_TYPE_KEY]:
            pass
        else:
            # Assume invariant
            if out_orig[k].dtype == torch.bool:
                err = (out_orig[k] != out_perm[k]).max()
            elif (out_orig[k].numel() == 0) and (out_perm[k].numel() == 0):
                err = 0.0
            else:
                err = (out_orig[k] - out_perm[k]).abs().max()
            fail = not torch.allclose(out_orig[k], out_perm[k], atol=atol)
            if fail:
                num_problems += 1
            messages.append(
                f"   edge & node permutation invariance for field {k:22} -> max error={err:.3e}{'  FAIL' if fail else ''}"
            )
    msg = "\n".join(messages)
    if num_problems == 0:
        return msg
    else:
        if raise_error:
            raise AssertionError(msg)
        else:
            return msg


def assert_AtomicData_equivariant(
    func: GraphModuleMixin,
    data_in: Union[AtomicDataDict.Type, List[AtomicDataDict.Type]],
    permutation_tolerance: Optional[float] = None,
    e3_tolerance: Optional[float] = None,
    **kwargs,
) -> str:
    r"""Test the rotation, translation, parity, and permutation equivariance of ``func``.

    For details on permutation testing, see ``assert_permutation_equivariant``.
    For details on geometric equivariance testing, see ``e3nn.util.test.assert_equivariant``.

    Raises ``AssertionError`` if issues are found.

    Args:
        func: the module or model to test
        data_in: the example input data(s) to test with. Only the first is used for permutation testing.
        **kwargs: passed to ``e3nn.util.test.assert_equivariant``

    Returns:
        A string description of the errors.
    """
    # Prevent pytest from showing this function in the traceback
    __tracebackhide__ = True

    if not isinstance(data_in, list):
        data_in = [data_in]
    data_in = [from_dict(d) for d in data_in]
    device, dtype = (
        data_in[0][AtomicDataDict.POSITIONS_KEY].device,
        data_in[0][AtomicDataDict.POSITIONS_KEY].dtype,
    )

    # == Test permutation of graph nodes ==
    # since permutation is discrete and should not be data dependent, run only on one frame.
    permutation_message = assert_permutation_equivariant(
        func, data_in[0], tolerance=permutation_tolerance, raise_error=False
    )

    # == Test rotation, parity, and translation using e3nn ==
    irreps_in = {k: None for k in AtomicDataDict.ALLOWED_KEYS}
    irreps_in.update(func.irreps_in)
    irreps_in = {k: v for k, v in irreps_in.items() if k in data_in[0]}
    irreps_out = func.irreps_out.copy()
    # Remove batch-related keys from the irreps_out, if we aren't using batched inputs
    irreps_out = {
        k: v
        for k, v in irreps_out.items()
        if not (k in ("batch", "ptr") and "batch" not in data_in)
    }
    # Assume that all data in `data_in` have the same keys
    # remove empty outputs from irreps out by running one set of data through
    data_in_0_copy = data_in[0].copy()
    ref_out = func(data_in_0_copy)
    for k, v in ref_out.items():
        if v.numel() == 0 and k in irreps_out.keys():
            _ = irreps_out.pop(k, None)

    # for certain things, we don't care what the given irreps are...
    # make sure that we test correctly for equivariance:
    for irps in (irreps_in, irreps_out):
        if AtomicDataDict.POSITIONS_KEY in irps:
            # it should always have been 1o vectors
            # since that's actually a valid Irreps
            assert o3.Irreps(irps[AtomicDataDict.POSITIONS_KEY]) == o3.Irreps("1o")
            irps[AtomicDataDict.POSITIONS_KEY] = "cartesian_points"
        if AtomicDataDict.CELL_KEY in irps:
            prev_cell_irps = irps[AtomicDataDict.CELL_KEY]
            assert prev_cell_irps is None or o3.Irreps(prev_cell_irps) == o3.Irreps(
                "1o"
            )
            # must be this to actually rotate it when flattened
            irps[AtomicDataDict.CELL_KEY] = "3x1o"

    cartesian_keys = _CARTESIAN_TENSOR_FIELDS.keys()
    for k in (
        AtomicDataDict.STRESS_KEY,
        AtomicDataDict.VIRIAL_KEY,
    ):  # TODO should this be cartesian_keys?
        irreps_in.pop(k, None)
    if any(k in irreps_out for k in cartesian_keys):
        from e3nn.io import CartesianTensor

        cartesian_tensor = {
            k: CartesianTensor(_CARTESIAN_TENSOR_FIELDS[k])
            for k in cartesian_keys
            if k in irreps_out
        }
        cartesian_rtp = {
            k: ct.reduced_tensor_products().to(device, dtype)
            for k, ct in cartesian_tensor.items()
        }
        for k, ct in cartesian_tensor.items():
            irreps_out[k] = ct

    def wrapper(*args):
        arg_dict = {k: v for k, v in zip(irreps_in, args)}
        # cell is a special case
        for key in (AtomicDataDict.CELL_KEY,):
            if key in arg_dict:
                # unflatten
                val = arg_dict[key]
                assert val.shape[-1] == 9
                arg_dict[key] = val.reshape(val.shape[:-1] + (3, 3))
        output = func(arg_dict)
        # irreps_out has been purged of numel=0 fields by now
        output = {k: output[k] for k, v in irreps_out.items() if k in output}
        # cell is a special case
        for key in (AtomicDataDict.CELL_KEY,):
            if key in output:
                # flatten
                val = output[key]
                assert val.shape[-2:] == (3, 3)
                output[key] = val.reshape(val.shape[:-2] + (9,))
        # cartesian tensors like stress are also a special case,
        # we need it to be decomposed into irreps for equivar testing
        for k in cartesian_keys:
            if k in output:
                output[k] = cartesian_tensor[k].from_cartesian(
                    output[k], rtp=cartesian_rtp[k].to(output[k].dtype)
                )
        output_list = []
        # TODO: this is a bit duct-taped and special cased for cell, but maybe it's ok?
        # potentially a more general solution is to inspect the irreps_out to set the dims of the dummy tensors accordingly
        # but the irreps_out values are strings (sometimes?)
        for k in irreps_out:
            if k in output:
                output_list.append(output[k])
            elif k == AtomicDataDict.CELL_KEY:
                # add a dummy cell
                output_list.append(torch.zeros(9, dtype=dtype, device=device))
            else:
                output_list.append(torch.zeros((1,), dtype=dtype, device=device))
        return output_list

    # prepare input data
    for d in data_in:
        # cell is a special case
        if AtomicDataDict.CELL_KEY in d:
            # flatten
            cell = d[AtomicDataDict.CELL_KEY]
            assert cell.shape[-2:] == (3, 3)
            d[AtomicDataDict.CELL_KEY] = cell.reshape(cell.shape[:-2] + (9,))

    errs = [
        equivariance_error(
            wrapper,
            args_in=[d[k] for k in irreps_in],
            irreps_in=list(irreps_in.values()),
            irreps_out=list(irreps_out.values()),
            **kwargs,
        )
        for d in data_in
    ]

    # take max across errors
    errs = {k: torch.max(torch.vstack([e[k] for e in errs]), dim=0)[0] for k in errs[0]}

    current_dtype = (
        func.model_dtype if isinstance(func, GraphModel) else torch.get_default_dtype()
    )
    if e3_tolerance is None:
        e3_tolerance = FLOAT_TOLERANCE[current_dtype]
    all_errs = []
    for case, err in errs.items():
        for key, this_err in zip(irreps_out.keys(), err):
            all_errs.append(case + (key, this_err))
    is_problem = [e[-1] > e3_tolerance for e in all_errs]

    message = (permutation_message + "\n") + "\n".join(
        f"   (parity_k={int(k[0]):1d}, did_translate={str(bool(k[1])):5}, field={str(k[2]):22})     -> max error={float(k[3]):.3e}{'  FAIL' if prob else ''}"
        for k, prob in zip(all_errs, is_problem)
        if irreps_out[str(k[2])] is not None
    )

    if any(is_problem) or " FAIL" in permutation_message:
        raise AssertionError(
            f"Equivariance test of {type(func).__name__} failed:\n   default dtype: {torch.get_default_dtype()} (assumed) model dtype: {current_dtype}  E(3) tolerance: {e3_tolerance}\n{message}"
        )

    return message


_DEBUG_HOOKS = None


def set_irreps_debug(enabled: bool = False) -> None:
    r"""Add debugging hooks to ``forward()`` that check data-irreps consistancy.

    Args:
        enabled: whether to set debug mode as enabled or disabled
    """
    global _DEBUG_HOOKS
    if _DEBUG_HOOKS is None and not enabled:
        return
    elif _DEBUG_HOOKS is not None and enabled:
        return
    elif _DEBUG_HOOKS is not None and not enabled:
        for hook in _DEBUG_HOOKS:
            hook.remove()
        _DEBUG_HOOKS = None
        return
    else:
        pass

    import torch.nn.modules

    def pre_hook(mod: GraphModuleMixin, inp):
        __tracebackhide__ = True
        if not isinstance(mod, GraphModuleMixin):
            return
        mname = type(mod).__name__
        if len(inp) > 1:
            raise ValueError(
                f"Module {mname} should have received a single argument, but got {len(inp)}"
            )
        elif len(inp) == 0:
            raise ValueError(
                f"Module {mname} didn't get any arguments; this case is correctly handled with an empty dict."
            )
        inp = inp[0]
        if not (isinstance(inp, dict)):
            raise TypeError(
                f"Module {mname} should have received a dict, instead got a {type(inp).__name__}"
            )
        for k, ir in mod.irreps_in.items():
            if k not in inp:
                pass
            elif isinstance(inp[k], torch.Tensor) and isinstance(ir, o3.Irreps):
                if inp[k].ndim == 1 and inp[k].numel() > 0:
                    raise ValueError(
                        f"Field {k} in input to module {mname} has only one dimension (assumed to be batch-like); it must have a second irreps dimension even if irreps.dim == 1 (i.e. a single per atom scalar must have shape [N_at, 1], not [N_at])"
                    )
                elif inp[k].shape[-1] != ir.dim and inp[k].numel() > 0:
                    raise ValueError(
                        f"Field {k} in input to module {mname} has last dimension {inp[k].shape[-1]} but its irreps {ir} indicate last dimension {ir.dim}"
                    )
        return

    h1 = torch.nn.modules.module.register_module_forward_pre_hook(pre_hook)

    def post_hook(mod: GraphModuleMixin, _, out):
        __tracebackhide__ = True
        if not isinstance(mod, GraphModuleMixin):
            return
        mname = type(mod).__name__
        if not (isinstance(out, dict)):
            raise TypeError(
                f"Module {mname} should have returned a dict, instead got a {type(out).__name__}"
            )
        for k, ir in mod.irreps_out.items():
            if k not in out:
                pass
            elif isinstance(out[k], torch.Tensor) and isinstance(ir, o3.Irreps):
                if out[k].ndim == 1 and out[k].numel() > 0:
                    raise ValueError(
                        f"Field {k} in output from module {mname} has only one dimension (assumed to be batch-like); it must have a second irreps dimension even if irreps.dim == 1 (i.e. a single per atom scalar must have shape [N_at, 1], not [N_at])"
                    )
                elif out[k].shape[-1] != ir.dim and out[k].numel() > 0:
                    raise ValueError(
                        f"Field {k} in output from {mname} has last dimension {out[k].shape[-1]} but its irreps {ir} indicate last dimension {ir.dim}"
                    )
        return

    h2 = torch.nn.modules.module.register_module_forward_hook(post_hook)

    _DEBUG_HOOKS = (h1, h2)
    return


def override_irreps_debug(enabled=True):
    """Decorator that toggles `set_irreps_debug` at the start of the test
    and restores it to the original state at the end.

    This decorator is crucial for PT2 compilation tests. (the hook modifications in `set_irreps_debug` may interfere with the compilation process).

    Args:
        enabled (bool): whether to enable or disable irreps debug mode
    """

    def decorator(test_func):
        @wraps(test_func)
        def wrapper(*args, **kwargs):
            original_state = _DEBUG_HOOKS is not None
            set_irreps_debug(enabled)
            try:
                return test_func(*args, **kwargs)
            finally:
                set_irreps_debug(original_state)

        return wrapper

    return decorator


def edgeset_from_AtomicDataDict(data, **nl_kwargs):
    data = compute_neighborlist_(data, **nl_kwargs)
    return set([tuple(edge) for edge in data["edge_index"].numpy().T])


def compare_neighborlists(
    atoms_or_data: Union[ase.Atoms, AtomicDataDict.Type],
    nl1: str,
    nl2: str,
    **nl_kwargs,
):
    """
    Args:
        nl1, nl2: the neighborlists to compare -- currently "ase", "matscipy", "vesin"
    """
    assert "r_max" in nl_kwargs
    assert "NL" not in nl_kwargs
    if isinstance(atoms_or_data, ase.Atoms):
        data = from_ase(atoms_or_data)
    else:
        data = atoms_or_data
    edges1 = edgeset_from_AtomicDataDict(data, NL=nl1, **nl_kwargs)
    edges2 = edgeset_from_AtomicDataDict(data, NL=nl1, **nl_kwargs)
    assert edges1 == edges2



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from .type_mapper import ChemicalSpeciesToAtomTypeMapper
from .neighborlist import NeighborListTransform, SortedNeighborListTransform
from .stress_utils import VirialToStressTransform, StressSignFlipTransform

__all__ = [
    "ChemicalSpeciesToAtomTypeMapper",
    "NeighborListTransform",
    "SortedNeighborListTransform",
    "VirialToStressTransform",
    "StressSignFlipTransform",
]



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from .utils import model_builder, override_model_compile_mode
from .modify_utils import modify
from .saved_models import ModelFromCheckpoint, ModelFromPackage
from .nequip_models import NequIPGNNModel, NequIPGNNEnergyModel, FullNequIPGNNModel
from .pair_potential import ZBLPairPotential

__all__ = [
    "model_builder",
    "override_model_compile_mode",
    "modify",
    "ModelFromCheckpoint",
    "ModelFromPackage",
    "NequIPGNNModel",
    "NequIPGNNEnergyModel",
    "FullNequIPGNNModel",
    "ZBLPairPotential",
]



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from ._graph_mixin import GraphModuleMixin, SequentialGraphNetwork
from .graph_model import GraphModel
from .atomwise import (
    AtomwiseOperation,
    AtomwiseReduce,
    AtomwiseLinear,
    PerTypeScaleShift,
)
from .nonlinearities import ShiftedSoftplus
from .mlp import ScalarMLP, ScalarMLPFunction
from .interaction_block import InteractionBlock
from .convnetlayer import ConvNetLayer
from .grad_output import GradientOutput, PartialForceOutput, ForceStressOutput
from .misc import Concat, ApplyFactor, SaveForOutput
from .utils import scatter, tp_path_exists, with_edge_vectors_
from .model_modifier_utils import model_modifier, replace_submodules

__all__ = [
    "GraphModel",
    "GraphModuleMixin",
    "SequentialGraphNetwork",
    "AtomwiseOperation",
    "AtomwiseReduce",
    "AtomwiseLinear",
    "PerTypeScaleShift",
    "ShiftedSoftplus",
    "ScalarMLP",
    "ScalarMLPFunction",
    "InteractionBlock",
    "GradientOutput",
    "PartialForceOutput",
    "ForceStressOutput",
    "ConvNetLayer",
    "Concat",
    "ApplyFactor",
    "SaveForOutput",
    "scatter",
    "tp_path_exists",
    "with_edge_vectors_",
    "model_modifier",
    "replace_submodules",
]



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.

from typing import List, Dict, Union


def per_edge_type_cutoff_to_metadata_str(
    type_names: List[str],
    per_edge_type_cutoff: Dict[str, Union[float, Dict[str, float]]],
    r_max: float,
) -> str:
    """Convert per-edge-type cutoff dict to flattened metadata string.

    Args:
        type_names: list of atom type names
        per_edge_type_cutoff: cutoff dict from config
        r_max: global cutoff radius

    Returns:
        space-separated string of cutoff values in row-major order
    """
    from ._edge import _process_per_edge_type_cutoff

    cutoff_tensor = _process_per_edge_type_cutoff(
        type_names, per_edge_type_cutoff, r_max
    )
    return " ".join(str(e.item()) for e in cutoff_tensor.view(-1))


def parse_per_edge_type_cutoff_metadata(
    cutoff_str: str, type_names: List[str]
) -> Dict[str, Dict[str, float]]:
    """Parse flattened per-edge-type cutoff metadata string to dict format.

    Reverse operation of `per_edge_type_cutoff_to_metadata_str`.

    Args:
        cutoff_str: space-separated string of cutoff values
        type_names: list of atom type names

    Returns:
        dict format suitable for NeighborListTransform
    """
    # short-circuit
    if cutoff_str in ("", None):
        return None

    cutoff_values = [float(x) for x in cutoff_str.split()]
    num_types = len(type_names)

    assert (
        len(cutoff_values) == num_types * num_types
    ), f"Expected {num_types * num_types} cutoff values, but got {len(cutoff_values)}"

    result = {}
    for i, source_type in enumerate(type_names):
        result[source_type] = {}
        for j, target_type in enumerate(type_names):
            result[source_type][target_type] = cutoff_values[i * num_types + j]

    return result



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import torch
import lightning
from lightning.pytorch.utilities.warnings import PossibleUserWarning
from hydra.utils import instantiate
from hydra.utils import get_method, get_class
from nequip.data import AtomicDataDict
from nequip.utils import RankedLogger

import warnings
from typing import Optional, Dict


logger = RankedLogger(__name__, rank_zero_only=True)


# metrics are already synced before logging, but Lightning still sends a PossibleUserWarning about setting sync_dist=True in self.logdict()
warnings.filterwarnings(
    "ignore",
    message=".*when logging on epoch level in distributed setting to accumulate the metric across.*",
    category=PossibleUserWarning,
)


_SOLE_MODEL_KEY = "sole_model"


class NequIPLightningModule(lightning.LightningModule):
    """:class:`~lightning.pytorch.core.LightningModule` for training, validating, testing and predicting with models constructed in the NequIP ecosystem.

    **Data**

    The ``NequIPLightningModule`` supports a single ``train`` dataset, but multiple ``val`` and ``test`` datasets.

    **Run Types and Metrics**

    - For ``train`` runs, users must provide ``loss`` and ``val_metrics``. The ``loss`` is computed on the training dataset to train the model, and requires each metric to have a corresponding coefficient that will be used to generate a ``weighted_sum``. This ``weighted_sum`` is the loss function that will be minimized over the course of training. ``val_metrics`` is computed on the validation dataset(s) for monitoring. Additionally, users may provide ``train_metrics`` to monitor metrics on the training dataset.
    - For ``val`` runs, users must provide ``val_metrics``.
    - For ``test`` runs, users must provide ``test_metrics``.

    **Logging Conventions**

    Logging is performed for the ``train``, ``val``, and ``test`` datasets.

    During ``train`` runs,
      * logging occurs at each batch ``step`` and at each ``epoch``,
      * there is only one training set, so no ``data_idx`` is used in the logging.

    For ``val`` and ``test`` runs,
      * logging only occurs at each validation or testing ``epoch``, i.e. one pass over the entirety of each validation/testing dataset,
      * there can be multiple validation and testing sets, so a zero-based ``data_idx`` index is used in the logging.

    Logging Format
      * ``/`` is used as a delimiter for to exploit the automatic grouping functionality of most loggers. Logged metrics will have the form ``train_{loss/metric}_{step/epoch}/{metric_name}`` and ``{val/test}{data_idx}_epoch/{metric_name}``. For example, ``train_loss_step/force_MSE``, ``train_metric_epoch/E_MAE``, ``val0_epoch/F_RMSE``, etc.
      * Note that this may have implications on how one would set the parameters for the `ModelCheckpoint <https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.callbacks.ModelCheckpoint.html>`_ callback, i.e. if the name of a metric is used in the checkpoint file's name, the ``/`` will cause a directory to be created when instead a file is desired.
    """

    def __init__(
        self,
        model: Dict,
        optimizer: Optional[Dict] = None,
        lr_scheduler: Optional[Dict] = None,
        loss: Optional[Dict] = None,
        train_metrics: Optional[Dict] = None,
        val_metrics: Optional[Dict] = None,
        test_metrics: Optional[Dict] = None,
        num_datasets: Optional[Dict[str, int]] = None,
        # for caching training info
        info_dict: Optional[Dict] = None,
    ):
        super().__init__()

        # save arguments to instantiate LightningModule from checkpoint automatically
        self.save_hyperparameters()

        # === instantiate model ===
        model_object = self._build_model(model)

        # === account for multiple models ===
        # contract:
        # - for multiple models, they must be in the form of a `ModuleDict` of `GraphModel`s
        # - if a single `GraphModel` is provided, we wrap it in a `ModuleDict`
        # - all models must have the same `type_names`

        # the reason for `hasattr(x, "is_graph_model")` and not just `isinstance(x, GraphModel)`
        # is to support `GraphModel` from a `nequip-package`d model (see https://pytorch.org/docs/stable/package.html#torch-package-sharp-edges)
        assert isinstance(model_object, torch.nn.ModuleDict) or hasattr(
            model_object, "is_graph_model"
        )
        if not isinstance(model_object, torch.nn.ModuleDict):
            model_object = torch.nn.ModuleDict({_SOLE_MODEL_KEY: model_object})
        self.model = model_object
        type_names_list = []
        for k, v in self.model.items():
            assert hasattr(v, "is_graph_model")
            type_names_list.append(v.type_names)
            logger.debug(f"Built Model Details ({k}):\n{str(v)}")
        assert all(
            [
                all(
                    [
                        name1 == name2
                        for (name1, name2) in zip(type_names_list[0], type_names)
                    ]
                )
                for type_names in type_names_list
            ]
        ), "If multiple models are used, they must have the same type names parameter."
        type_names = type_names_list[0]  # passed to `MetricsManager`s later

        # === optimizer and lr scheduler ===
        self.optimizer_config = optimizer
        self.lr_scheduler_config = lr_scheduler

        # === instantiate MetricsManager objects ===
        # must have separate MetricsManagers for each dataloader
        # num_datasets goes in order [train, val, test, predict]
        self.num_datasets = (
            num_datasets
            if num_datasets is not None
            else {
                "train": 0,
                "val": 0,
                "test": 0,
                "predict": 0,
            }
        )

        assert (
            self.num_datasets["train"] == 1
        ), "currently only support one training dataset"

        # == DDP concerns for loss ==

        # to account for loss contributions from multiple ranks later on
        # NOTE: this must be updated externally by the script that sets up the training run
        self.world_size = 1

        # == instantiate loss ==
        self.loss = instantiate(loss, type_names=type_names)
        if self.loss is not None:
            assert (
                self.loss.do_weighted_sum
            ), "`coeff` must be set for entries of the `loss` MetricsManager for a weighted sum of metrics components to be used as the loss."

        # set `dist_sync_on_step=True` for loss metrics
        # to ensure correct DDP syncing of loss function for batch steps
        for metric in self.loss.values():
            metric.dist_sync_on_step = True
        self.loss.eval()

        # == instantiate other metrics ==
        self.train_metrics = instantiate(train_metrics, type_names=type_names)
        if self.train_metrics is not None:
            self.train_metrics.eval()
        # may need to instantate multiple instances to account for multiple val and test datasets
        self.val_metrics = torch.nn.ModuleList(
            [
                instantiate(val_metrics, type_names=type_names)
                for _ in range(self.num_datasets["val"])
            ]
        )
        if self.val_metrics is not None:
            self.val_metrics.eval()
        self.test_metrics = torch.nn.ModuleList(
            [
                instantiate(test_metrics, type_names=type_names)
                for _ in range(self.num_datasets["test"])
            ]
        )
        if self.test_metrics is not None:
            self.test_metrics.eval()

        # use "/" as delimiter for loggers to automatically categorize logged metrics
        self.logging_delimiter = "/"

        # for statefulness of the run stage
        self.register_buffer("run_stage", torch.zeros((1), dtype=torch.long))

    def _build_model(self, model_config: Dict) -> torch.nn.ModuleDict:
        """Constructs a ``torch.nn.ModuleDict[str, nequip.nn.GraphModel]`` from a pure Python dictionary.

        Subclasses that require more control over how the model is built can override this method.
        """
        # reason for following implementation instead of just `hydra.utils.instantiate(model)` is to prevent omegaconf from being a model dependency
        model_config = model_config.copy()  # make a copy because of `pop` mutation
        model_builder = get_method(model_config.pop("_target_"))
        model = model_builder(**model_config)
        return model

    def configure_optimizers(self):
        """"""
        # currently support 1 optimizer and 1 scheduler
        # potentially support N optimzier and N scheduler
        # (see https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.core.LightningModule.html#lightning.pytorch.core.LightningModule.configure_optimizers)
        optimizer_config = self.optimizer_config.copy()
        param_groups = optimizer_config.pop(
            "param_groups",
            {"_target_": "nequip.train.lightning._default_param_group_factory"},
        )
        param_groups = instantiate(param_groups, model=self.model)
        optimizer_class = optimizer_config.pop("_target_")
        optim = get_class(optimizer_class)(params=param_groups, **optimizer_config)
        if self.lr_scheduler_config is not None:
            # instantiate lr scheduler object separately to pass the optimizer to it during instantiation
            lr_scheduler_config = dict(self.lr_scheduler_config.copy())
            scheduler = lr_scheduler_config.pop("scheduler")
            scheduler = instantiate(scheduler, optimizer=optim)
            lr_scheduler = dict(instantiate(lr_scheduler_config))
            lr_scheduler.update({"scheduler": scheduler})
            return {"optimizer": optim, "lr_scheduler": lr_scheduler}

        return optim

    def forward(self, inputs: AtomicDataDict.Type):
        """"""
        # enable grad for forces, stress, etc
        with torch.enable_grad():
            # multi-model subclasses will need to override this function
            return self.model[_SOLE_MODEL_KEY](inputs)

    @property
    def evaluation_model(self) -> torch.nn.Module:
        return self.model

    def process_target(
        self, batch: AtomicDataDict.Type, batch_idx: int, dataloader_idx: int = 0
    ) -> AtomicDataDict.Type:
        """"""
        # subclasses can override this function
        return batch.copy()

    def training_step(
        self, batch: AtomicDataDict.Type, batch_idx: int, dataloader_idx: int = 0
    ):
        """"""
        target = self.process_target(batch, batch_idx, dataloader_idx)
        output = self(batch)

        # optionally compute training metrics
        if self.train_metrics is not None:
            with torch.no_grad():
                train_metric_dict = self.train_metrics(
                    output, target, prefix=f"train_metric_step{self.logging_delimiter}"
                )
            self.log_dict(train_metric_dict)

        # compute loss and return
        loss_dict = self.loss(
            output, target, prefix=f"train_loss_step{self.logging_delimiter}"
        )
        self.log_dict(loss_dict)
        # In DDP training, because gradients are averaged rather than summed over nodes,
        # we get an effective factor of 1/n_rank applied to the loss. Because our loss already
        # manages correct accumulation of the metric over ranks, we want to cancel out this
        # unnecessary 1/n_rank term. If DDP is disabled, this is 1 and has no effect.
        loss = (
            loss_dict[f"train_loss_step{self.logging_delimiter}weighted_sum"]
            * self.world_size
        )
        return loss

    def on_train_epoch_end(self):
        """"""
        # optionally compute training metrics
        if self.train_metrics is not None:
            train_metric_dict = self.train_metrics.compute(
                prefix=f"train_metric_epoch{self.logging_delimiter}"
            )
            self.log_dict(train_metric_dict)
            self.train_metrics.reset()
        # loss
        loss_dict = self.loss.compute(
            prefix=f"train_loss_epoch{self.logging_delimiter}"
        )
        self.log_dict(loss_dict)
        self.loss.reset()

    def validation_step(
        self, batch: AtomicDataDict.Type, batch_idx: int, dataloader_idx: int = 0
    ):
        """"""
        target = self.process_target(batch, batch_idx, dataloader_idx)

        # === update basic val metrics ===
        output = self(batch)
        with torch.no_grad():
            metric_dict = self.val_metrics[dataloader_idx](
                output,
                target,
                prefix=f"val{dataloader_idx}_step{self.logging_delimiter}",
            )

        return metric_dict

    def on_validation_epoch_end(self):
        """"""
        # === reset basic val metrics ===
        for idx, metrics in enumerate(self.val_metrics):
            metric_dict = metrics.compute(
                prefix=f"val{idx}_epoch{self.logging_delimiter}"
            )
            self.log_dict(metric_dict)
            metrics.reset()

    def test_step(
        self, batch: AtomicDataDict.Type, batch_idx: int, dataloader_idx: int = 0
    ):
        """"""
        target = self.process_target(batch, batch_idx, dataloader_idx)

        # === update basic test metrics ===
        output = self(batch)
        with torch.no_grad():
            metric_dict = self.test_metrics[dataloader_idx](
                output,
                target,
                prefix=f"test{dataloader_idx}_step{self.logging_delimiter}",
            )
        metric_dict.update({f"test_{dataloader_idx}_output": output})

        return metric_dict

    def on_test_epoch_end(self):
        """"""
        # === reset basic test metrics ===
        for idx, metrics in enumerate(self.test_metrics):
            metric_dict = metrics.compute(
                prefix=f"test{idx}_epoch{self.logging_delimiter}"
            )
            self.log_dict(metric_dict)
            metrics.reset()


def _default_param_group_factory(model):
    return model.parameters()



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from nequip.data.datamodule import ASEDataModule
from nequip.utils import download_url, extract_zip
from nequip.utils.logger import RankedLogger

import os
from typing import Union, Sequence, List, Callable

logger = RankedLogger(__name__, rank_zero_only=True)

_URL_TM23 = "https://archive.materialscloud.org/record/file?record_id=2113&filename=benchmarking_master_collection-20240316T202423Z-001.zip"
supported_elements = [
    "Ag",
    "Au",
    "Cd",
    "Co",
    "Cr",
    "Cu",
    "Fe",
    "Hf",
    "Hg",
    "Ir",
    "Mn",
    "Mo",
    "Nb",
    "Ni",
    "Os",
    "Pd",
    "Pt",
    "Re",
    "Rh",
    "Ta",
    "Tc",
    "Ti",
    "V",
    "W",
    "Zn",
    "Zr",
]


class TM23DataModule(ASEDataModule):
    """LightningDataModule for the `TM23 dataset <https://www.nature.com/articles/s41524-024-01264-z>`_.

    This datamodule can be used for ``train``, ``validate``, and ``test`` runs.

    This datamodule can automatically download the TM23 dataset from https://archive.materialscloud.org/record/2024.48
    and unzip it in ``data_source_dir`` if not already downloaded. Otherwise, one can download and unzip the dataset as is and
    set ``data_source_dir`` to the directory that contains ``benchmarking_master_collection``.

    The combined dataset containing cold, warm, and melt frames are used as the train and test datasets. ``element`` can be any
    TM23 element, including ``Ag``, ``Au``, ``Cd``, ``Co``, ``Cr``, ``Cu``, ``Fe``, ``Hf``, ``Hg``, ``Ir``, ``Mn``, ``Mo``,
    ``Nb``, ``Ni``, ``Os``, ``Pd``, ``Pt``, ``Re``, ``Rh``, ``Ta``, ``Tc``, ``Ti``, ``V``, ``W``, ``Zn``, and ``Zr``.

    The ``train_val_split`` argument is required to split the training dataset into separate training and validation datasets.

    Args:
        seed (int): data seed for reproducibility
        data_source_dir (str): directory containing the TM23 dataset if present, else directory where TM23 dataset will be downloaded to
        element(str): element from TM23 dataset to use
        transforms (List[Callable]): list of data transforms
        train_val_split (List[float] or List[int]): train-validation split either in fractions ``[1, 1-f]`` or integers ``[N_train, N_val]``
    """

    def __init__(
        self,
        seed: int,
        data_source_dir: str,
        element: str,
        transforms: List[Callable],
        train_val_split: Sequence[Union[int, float]],
        **kwargs,
    ):
        assert element in supported_elements

        train_file_path = "/".join(
            [
                data_source_dir,
                "benchmarking_master_collection",
                element + "_2700cwm_train.xyz",
            ]
        )
        test_file_path = "/".join(
            [
                data_source_dir,
                "benchmarking_master_collection",
                element + "_2700cwm_test.xyz",
            ]
        )
        super().__init__(
            seed=seed,
            test_file_path=test_file_path,
            split_dataset={
                "file_path": train_file_path,
                "train": train_val_split[0],
                "val": train_val_split[1],
            },
            transforms=transforms,
            ase_args={"format": "extxyz"},
            **kwargs,
        )
        self.element = element
        self.data_source_dir = data_source_dir
        self.train_file_path = train_file_path
        self.test_file_path = test_file_path

    def prepare_data(self):
        """"""
        if not (
            os.path.isfile(self.test_file_path) and os.path.isfile(self.train_file_path)
        ):
            download_path = download_url(_URL_TM23, self.data_source_dir)
            extract_zip(download_path, self.data_source_dir)
        else:
            logger.info(
                f"Using existing data files `{self.train_file_path}` and `{self.test_file_path}`"
            )



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import lightning
from lightning.pytorch.callbacks import Callback
from nequip.train import NequIPLightningModule
from typing import Dict


class LossCoefficientScheduler(Callback):
    """Schedule loss coefficients during training.

    The ``LossCoefficientScheduler`` takes a single argument ``schedule``, which is a ``Dict[int, Dict[str, float]]`` where the keys are the epochs at which the loss coefficients change and the values are dictionaries mapping loss metric names (corresponding to how the loss was configured) to their coefficients.

    When the trainer's epoch counter matches any of the keys (representing epochs), the loss coefficients will be changed to the values (representing the coefficients for each loss term).

    The coefficients will be normalized to sum up to 1 in line with the convention of :class:`~nequip.train.MetricsManager`.

    Example usage in config where there are two loss contributions:

    .. code-block:: yaml

        callbacks:
          - _target_: nequip.train.callbacks.LossCoefficientScheduler
            schedule:
              100:
                per_atom_energy_mse: 1.0
                forces_mse: 5.0
              200:
                per_atom_energy_mse: 5.0
                forces_mse: 1.0

    Args:
        schedule (Dict[int, Dict[str,float]]): map of epoch to loss coefficient dictionary
    """

    def __init__(self, schedule: Dict[int, Dict[str, float]]):
        # ensure that the keys are `int`s
        self.schedule = {int(k): v for k, v in schedule.items()}
        # sanity check - epochs are >= 0
        assert all([epoch >= 0 for epoch in self.schedule.keys()])

    def on_train_epoch_start(
        self,
        trainer: lightning.Trainer,
        pl_module: NequIPLightningModule,
    ) -> None:
        """"""
        # only change loss coefficients at the designated epochs
        if trainer.current_epoch not in self.schedule.keys():
            return
        # set the loss coefficients
        pl_module.loss.set_coeffs(self.schedule[trainer.current_epoch])



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from ._key_registry import (
    register_fields,
    deregister_fields,
    _register_field_prefix,
    get_field_type,
    _NODE_FIELDS,
    _EDGE_FIELDS,
    _GRAPH_FIELDS,
    _LONG_FIELDS,
    _CARTESIAN_TENSOR_FIELDS,
    ABBREV,
)
from .dict import from_dict
from .ase import from_ase, to_ase
from ._nl import compute_neighborlist_
from ._sampler import PartialSampler
from .stats import (
    Count,
    Mean,
    MeanAbsolute,
    RootMeanSquare,
    StandardDeviation,
    Max,
    Min,
)
from .stats_manager import (
    DataStatisticsManager,
    CommonDataStatisticsManager,
    EnergyOnlyDataStatisticsManager,
)
from .modifier import BaseModifier, PerAtomModifier, EdgeLengths, NumNeighbors


__all__ = [
    "register_fields",
    "deregister_fields",
    "_register_field_prefix",
    "get_field_type",
    "PartialSampler",
    "_NODE_FIELDS",
    "_EDGE_FIELDS",
    "_GRAPH_FIELDS",
    "_LONG_FIELDS",
    "_CARTESIAN_TENSOR_FIELDS",
    "ABBREV",
    "from_dict",
    "from_ase",
    "to_ase",
    "compute_neighborlist_",
    "DataStatisticsManager",
    "CommonDataStatisticsManager",
    "EnergyOnlyDataStatisticsManager",
    "Count",
    "Mean",
    "MeanAbsolute",
    "RootMeanSquare",
    "StandardDeviation",
    "Max",
    "Min",
    "BaseModifier",
    "PerAtomModifier",
    "EdgeLengths",
    "NumNeighbors",
]



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import torch

import math

# Technically, use of this module should probably be guarded by conditional_torchscript_jit
# But its use as a drop-in replacement for functions like torch.nn.functional.silu makes that
# difficult, so, given the rarety of its use, we have just removed @torch.jit.script


def shifted_softplus(x):
    return torch.nn.functional.softplus(x) - math.log(2.0)


class ShiftedSoftplus(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return shifted_softplus(x)



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from .base_datasets import AtomicDataset
from .lmdb_dataset import NequIPLMDBDataset
from .ase_dataset import ASEDataset
from .npz_dataset import NPZDataset
from .hdf5_dataset import HDF5Dataset
from .test_data import EMTTestDataset
from .utils import SubsetByRandomSlice, RandomSplitAndIndexDataset


__all__ = [
    "AtomicDataset",
    "NequIPLMDBDataset",
    "ASEDataset",
    "NPZDataset",
    "HDF5Dataset",
    "EMTTestDataset",
    "SubsetByRandomSlice",
    "RandomSplitAndIndexDataset",
]



from ._tp_scatter_base import TensorProductScatter


class OpenEquivarianceTensorProductScatter(TensorProductScatter):

    def __init__(
        self,
        feature_irreps_in,
        irreps_edge_attr,
        irreps_mid,
        instructions,
        use_opaque: bool,
    ) -> None:
        super().__init__(
            feature_irreps_in=feature_irreps_in,
            irreps_edge_attr=irreps_edge_attr,
            irreps_mid=irreps_mid,
            instructions=instructions,
        )
        # ^ we ensure that the base class keeps around a `self.tp` that carries its own set of persistent buffers
        # even though `self.tp` is not used, having its (persistent) buffers always around ensures state dict compatibility when adding on or removing this subclass module

        # === OEQ ===

        # we do lazy imports of oeq to allow `nequip-package` to pick this file up even if oeq is not installed
        # since `nequip-package` ignores files if it errors on loading the file

        from openequivariance import (
            TensorProductConv,
            TPProblem,
            torch_to_oeq_dtype,
        )

        tpp = TPProblem(
            feature_irreps_in,
            irreps_edge_attr,
            irreps_mid,
            instructions,
            irrep_dtype=torch_to_oeq_dtype(self.model_dtype),
            weight_dtype=torch_to_oeq_dtype(self.model_dtype),
            shared_weights=False,
            internal_weights=False,
        )
        self.tp_conv = TensorProductConv(
            tpp, torch_op=True, deterministic=False, use_opaque=use_opaque
        )

    def forward(self, x, edge_attr, edge_weight, edge_dst, edge_src):
        # explicit cast to account for AMP
        return self.tp_conv(
            x.to(self.model_dtype),
            edge_attr.to(self.model_dtype),
            edge_weight.to(self.model_dtype),
            edge_dst,
            edge_src,
        )



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import torch
from torchmetrics import Metric
from nequip.data import AtomicDataDict, BaseModifier, PerAtomModifier
from .metrics import (
    MeanSquaredError,
    MeanAbsoluteError,
    RootMeanSquaredError,
)

from typing import List, Dict, Union, Callable, Final, Optional

_METRICS_MANAGER_INPUT_KEYS: Final[List[str]] = [
    "field",
    "metric",
    "name",
    "coeff",
    "per_type",
    "ignore_nan",
]

_METRICS_MANAGER_MANDATORY_INPUT_KEYS: Final[List[str]] = [
    "metric",
]


class MetricsManager(torch.nn.ModuleDict):
    """Manages ``nequip`` metrics that can be applied to ``AtomicDataDict`` to compute error metrics.

    This class manages both metrics for loss functions and metrics for monitoring and reporting purposes. The main input argument ``metrics`` is a list of dictionaries, where each dictionary contains the following keys.

    There are two mandatory keys:

    - ``field`` refers to the quantity of interest for metric computation. It has two formats:

      - a ``str`` for a ``nequip`` defined field (e.g. ``total_energy``, ``forces``, ``stress``), or
      - a ``Callable`` that performs some additional operations before returning a :class:`torch.Tensor`
        for metric computation (e.g. :class:`~nequip.data.PerAtomModifier`).

    - ``metric`` is a :class:`torchmetrics.Metric`. Users are expected to mostly use :class:`~nequip.train.MeanSquaredError` and :class:`~nequip.train.MeanAbsoluteError` for MSEs (for loss), RMSEs, and MAEs (for monitoring).

    The remaining keys are optional:

    - ``per_type`` is a ``bool`` (defaults to ``False`` if not provided). If ``True``, node fields (such as ``forces``) will have their metrics computed separately for each atom type based on the ``type_names`` argument. A simple average over the per-type metrics will be used as the "effective" metric returned. There are some subtleties:

      1) During batch steps, the per-type metric for a particular type may be ``NaN`` (by design) if the batch does not contain that atom type. Correspondingly, the per-batch effective metric does not account for that particular type, and is only averaged over the number of atom types that were in that batch.
      2) The per-epoch effective metric will always consider all atom types configured in its computation (i.e. when taking the simple average) since all atom types are expected to have contributed to at least one batch step over an epoch.

    - ``coeff`` is a ``float`` that determines the relative weight of the metric contribution to an overall ``weighted_sum`` of metrics. A ``weighted_sum`` is automatically computed if any input dictionary contains a float-valued ``coeff``. This feature is important for loss functions for multitask problems (e.g. training on ``total_energy`` and ``forces`` simultaneously), and for constructing an effective validation metrics for monitoring (e.g. for early stopping and lr scheduling). Entries without ``coeff`` will still have their metrics computed, but they will not be incorporated into the ``weighted_sum``. An example for the utility of this feature is for monitoring additional metrics of the same nature (e.g. energy MSE and MAE) but only wanting one of them be in the effective metric used for lr scheduling. Note that these coefficients will be **automatically normalized** to sum up to one, e.g. if one has an energy metric with ``coeff=3`` and a force metric with ``coeff=1``, the ``weighted_sum`` is computed with coefficients ``[0.75, 0.25]`` for the energy and force metric respectively.

    - ``ignore_nan`` is a ``bool`` (defaults to ``False`` if not provided). This should be set to true if one expects the underlying ``target`` data to contain ``NaN`` entries. An example use case is when one has a dataset with ``stress`` labels for only a portion of the dataset. One can still train on ``stress`` for data that contain it and the others can be set as ``NaN`` entries to be handled appropriately during metric computation with this key.

    - ``name`` is the name that the metric is logged as. Default names are used if not provided, but it is recommended for users to set custom names for clarity and control.

    Args:
        metrics (list): list of dictionaries with keys ``field``, ``metric``, ``per_type``, ``coeff``, ``ignore_nan``, and ``name``
        type_names (list): required for ``per_type`` metrics (if this class is instantiated in :class:`~nequip.train.NequIPLightningModule`, which is the case if one uses ``nequip-train``, this is automatically handled such that users need not explicitly fill in this field in the config)
    """

    def __init__(
        self,
        metrics: List[
            Dict[str, Union[float, str, Dict[str, Union[str, Callable]], Metric]]
        ],
        type_names: List[str] = None,
    ):
        super().__init__()
        self.metrics: Dict[str, Dict[str, Union[float, str, bool]]] = {}
        per_type_encountered = False
        for metric_dict in metrics:
            # === sanity checks ===
            # == check that keys present are expected ==
            for key in metric_dict.keys():
                assert (
                    key in _METRICS_MANAGER_INPUT_KEYS
                ), f"unrecognized key `{key}` found as input in `MetricsManager`"

            # == check that mandatory keys are present ==
            for mandatory_key in _METRICS_MANAGER_MANDATORY_INPUT_KEYS:
                assert (
                    mandatory_key in metric_dict.keys()
                ), f"Each dictionary in `MetricManager`'s `metrics` argument must possess at least the following mandatory keys: {_METRICS_MANAGER_MANDATORY_INPUT_KEYS}"

            # === field ===
            # field can be
            # - str: we will wrap it with `BaseModifier`
            # - Callable: it should be a subclass of `BaseModifier`
            # - None: this is a special metric that will have a different code path than the rest
            field = metric_dict.get("field", None)
            field = BaseModifier(field) if isinstance(field, str) else field
            if field is not None:
                assert isinstance(field, BaseModifier)

            # === metric ===
            metric = metric_dict["metric"]

            # === names ===
            name = metric_dict.get("name", None)
            if name is None:
                name = (
                    "_".join([str(field), str(metric)])
                    if field is not None
                    else str(metric)
                )
            assert (
                name != "weighted_sum"
            ), "`weighted_sum` is a specially reserved metric name that should not be configured."
            assert (
                name not in self.metrics.keys()
            ), f"Repeated names found ({name}) -- names must be unique. It is recommended to give custom names instead of relying on the automatic naming."

            # === sanity check `field=None` case ===
            if field is None:
                for key in ("ignore_nan", "per_type"):
                    assert (
                        key not in metric_dict
                    ), f"When field is not provided or `field: None`, `{key}` should not be provided."
                # NOTE that we still go through the `ignore_nan` and `per_type` handling below so they get the default values
                # so special metrics will have to handle NaNs on their own, and cannot be used with the automatic `per_type` system

            # == ignore Nan ==
            ignore_nan = metric_dict.get("ignore_nan", False)
            assert isinstance(
                ignore_nan, bool
            ), f"`ignore_nan` should be a bool, but found {ignore_nan} of type {type(ignore_nan)}"

            # == per_type metrics ==
            per_type = metric_dict.get("per_type", False)
            # sanity check that `type_names` is provided
            # to only do the assert once
            if per_type and not per_type_encountered:
                per_type_encountered = True
                assert (
                    type_names is not None
                ), "`type_names` must be provided if any `per_type=True`"
                self.type_names = type_names
            # logic to proliferate metrics objects
            if per_type:
                if field.type != "node":
                    raise RuntimeError(
                        f"`per_type` metrics only supported for node fields, but {field.type} field found for {name}."
                    )
                # set up per_type metrics as a ModuleList
                # one copy of the base Metric for each type in forward() and compute()
                ptm_list = torch.nn.ModuleList([])
                for _ in range(len(self.type_names)):
                    ptm_list.append(metric.clone())
                metric = ptm_list

            # === construct dict entry ===
            self.metrics.update(
                {
                    name: {
                        "field": field,
                        "coeff": metric_dict.get("coeff", None),
                        "ignore_nan": ignore_nan,
                        "per_type": per_type,
                    }
                }
            )
            self.update({name: metric})

            # == clean up for safety ==
            if field is not None:
                del ignore_nan, per_type
            del field, metric, name

        # === process coefficients ===
        self.do_weighted_sum = False
        self.set_coeffs(
            {k: v["coeff"] for k, v in self.metrics.items()}
        )  # normalize coefficients if not all None

        # convenient to cache metrics computed last for callbacks
        self.metrics_values_step = {k: None for k in self.metrics.keys()}
        self.metrics_values_epoch = {k: None for k in self.metrics.keys()}

    def forward(
        self,
        preds: AtomicDataDict.Type,
        target: AtomicDataDict.Type,
        prefix: str = "",
        suffix: str = "",
    ):
        """
        Computes and accumulates metrics (intended for use at batch steps).
        """
        self.metrics_values_step = {k: None for k in self.metrics.keys()}
        if self.do_weighted_sum:
            weighted_sum = 0.0
        metric_dict = {}

        for metric_name, metric_params in self.metrics.items():
            field: Optional[Callable] = metric_params["field"]

            if field is not None:
                per_type: bool = metric_params["per_type"]
                ignore_nan: bool = metric_params["ignore_nan"]

                preds_field, target_field = field(preds, target)
                if per_type:
                    metric = 0
                    num_contributing_types = 0
                    for type_idx, type_name in enumerate(self.type_names):
                        # index out each type
                        selector = torch.eq(
                            preds[AtomicDataDict.ATOM_TYPE_KEY], type_idx
                        )
                        per_type_preds = preds_field[selector]
                        per_type_target = target_field[selector]

                        # mask out NaNs (based on target)
                        if ignore_nan:
                            notnan_mask = ~torch.isnan(per_type_target)
                            per_type_preds = torch.masked_select(
                                per_type_preds, notnan_mask
                            )
                            per_type_target = torch.masked_select(
                                per_type_target, notnan_mask
                            )

                        pt_metric = self[metric_name][type_idx](
                            per_type_preds, per_type_target
                        )
                        pt_metric_name = (
                            prefix + "_".join([metric_name, type_name]) + suffix
                        )
                        metric_dict.update({pt_metric_name: pt_metric})
                        # account for batches without atom type
                        assert pt_metric.numel() == 1
                        if not torch.isnan(pt_metric):
                            metric = metric + pt_metric
                            num_contributing_types += 1
                    assert num_contributing_types <= len(self.type_names)
                    metric = metric / num_contributing_types
                else:
                    # mask out NaNs (based on target)
                    if ignore_nan:
                        notnan_mask = ~torch.isnan(target_field)
                        preds_field = torch.masked_select(preds_field, notnan_mask)
                        target_field = torch.masked_select(target_field, notnan_mask)
                    metric = self[metric_name](preds_field, target_field)
            else:
                # custom metric that takes in two AtomicDataDict objects
                metric = self[metric_name](preds, target)

            metric_dict.update({prefix + metric_name + suffix: metric})
            self.metrics_values_step.update({metric_name: metric.item()})

            if self.do_weighted_sum:
                coeff: Optional[float] = metric_params["coeff"]
                if coeff is not None:
                    weighted_sum = weighted_sum + metric * coeff

        if self.do_weighted_sum:
            metric_dict.update({prefix + "weighted_sum" + suffix: weighted_sum})

        return metric_dict

    def compute(self, prefix: str = "", suffix: str = ""):
        """
        Aggregates accumulated metrics (intended for use at the end of an epoch).
        """
        self.metrics_values_epoch = {k: None for k in self.metrics.keys()}
        if self.do_weighted_sum:
            weighted_sum = 0.0
        metric_dict = {}
        for metric_name, metric_params in self.metrics.items():
            if metric_params["per_type"]:
                metric = 0
                for type_idx, type_name in enumerate(self.type_names):
                    ps_metric = self[metric_name][type_idx].compute()
                    ps_metric_name = (
                        prefix + "_".join([metric_name, type_name]) + suffix
                    )
                    metric_dict.update({ps_metric_name: ps_metric})
                    metric = metric + ps_metric
                metric = metric / len(self.type_names)
            else:
                metric = self[metric_name].compute()
            metric_dict.update({prefix + metric_name + suffix: metric})
            self.metrics_values_epoch.update({metric_name: metric.item()})
            if self.do_weighted_sum:
                coeff = metric_params["coeff"]
                if coeff is not None:
                    weighted_sum = weighted_sum + metric * coeff
        if self.do_weighted_sum:
            metric_dict.update({prefix + "weighted_sum" + suffix: weighted_sum})

        return metric_dict

    def reset(self):
        for metric_name, metric_params in self.metrics.items():
            if metric_params["per_type"]:
                for type_idx in range(len(self.type_names)):
                    self[metric_name][type_idx].reset()
            else:
                self[metric_name].reset()

    def set_coeffs(self, coeff_dict: Dict[str, Optional[float]]) -> None:
        """
        Sanity checks and normalizes coefficients to one before setting the new coefficients.
        If some metrics are unspecified, the ``coeff`` will be assumed to be ``None``.
        """
        # add `None` if metric coeff is unspecified
        for metric_name in self.metrics.keys():
            if metric_name not in coeff_dict.keys():
                coeff_dict[metric_name] = None
        # normalize the coeffs
        if not all([v is None for k, v in coeff_dict.items()]):
            # normalize coefficients to sum up to 1 wherever provided
            tot = sum([v if v is not None else 0.0 for k, v in coeff_dict.items()])
            for name, metric_dict in self.metrics.items():
                metric_dict["coeff"] = (
                    coeff_dict[name] / tot if coeff_dict[name] is not None else None
                )
            self.do_weighted_sum = True
        else:
            for name, metric_dict in self.metrics.items():
                metric_dict["coeff"] = None
            self.do_weighted_sum = False

    def get_extra_state(self) -> None:
        """"""
        return {
            "coeff_dict": {k: v["coeff"] for k, v in self.metrics.items()},
            "metrics_values_step": self.metrics_values_step,
            "metrics_values_epoch": self.metrics_values_epoch,
        }

    def set_extra_state(self, state: Dict) -> None:
        """"""
        self.set_coeffs(state["coeff_dict"])
        self.metrics_values_step = state["metrics_values_step"]
        self.metrics_values_epoch = state["metrics_values_epoch"]


def EnergyForceLoss(
    coeffs: Dict[str, float] = {
        AtomicDataDict.TOTAL_ENERGY_KEY: 1.0,
        AtomicDataDict.FORCE_KEY: 1.0,
    },
    per_atom_energy: bool = True,
    type_names=None,
):
    """Simplified :class:`MetricsManager` wrapper for a **loss** term containing energy and forces mean squared errors (MSEs).

    The loss component names are ``per_atom_energy_mse`` OR ``total_energy_mse`` (depending on whether ``per_atom_energy`` is ``True`` or ``False``), and ``forces_mse``, which are the names to refer to when neeeded, e.g. when scheduling loss component coefficients.

    Example usage in config:

    .. code-block:: yaml

        training_module:
          _target_: nequip.train.NequIPLightningModule

          loss:
            _target_: nequip.train.EnergyForceLoss
            per_atom_energy: true
            coeffs:
              total_energy: 1.0
              forces: 1.0

    Args:
        coeffs (Dict[str, float]): ``dict`` that stores the relative weight of energy and forces to the overall loss (default ``{'total_energy': 1.0, 'forces': 1.0}``)
        per_atom_energy (bool, optional): whether to normalize the total energy by the number of atoms (default ``True``)
    """

    metrics = [
        {
            "name": "per_atom_energy_mse" if per_atom_energy else "total_energy_mse",
            "field": (
                PerAtomModifier(AtomicDataDict.TOTAL_ENERGY_KEY)
                if per_atom_energy
                else AtomicDataDict.TOTAL_ENERGY_KEY
            ),
            "coeff": coeffs[AtomicDataDict.TOTAL_ENERGY_KEY],
            "metric": MeanSquaredError(),
        },
        {
            "name": "forces_mse",
            "field": AtomicDataDict.FORCE_KEY,
            "coeff": coeffs[AtomicDataDict.FORCE_KEY],
            "metric": MeanSquaredError(),
        },
    ]
    return MetricsManager(metrics, type_names=type_names)


_EF_METRICS_COEFFS_KEYS: Final[List[str]] = [
    "total_energy_rmse",
    "per_atom_energy_rmse",
    "forces_rmse",
    "total_energy_mae",
    "per_atom_energy_mae",
    "forces_mae",
]

_EFS_METRICS_COEFFS_KEYS: Final[List[str]] = _EF_METRICS_COEFFS_KEYS + [
    "stress_rmse",
    "stress_mae",
]


def EnergyForceMetrics(
    coeffs: Dict[str, float] = {
        "total_energy_rmse": 1.0,
        "per_atom_energy_rmse": None,
        "forces_rmse": 1.0,
        "total_energy_mae": None,
        "per_atom_energy_mae": None,
        "forces_mae": None,
    },
    type_names=None,
):
    """Simplified :class:`MetricsManager` wrapper for a **metric** term containing energy and force mean absolute errors (MAEs) and root mean squared errors (RMSEs).

    Example usage in config:

    .. code-block:: yaml

        training_module:
          _target_: nequip.train.NequIPLightningModule

          val_metrics:
            _target_: nequip.train.EnergyForceMetrics
            coeffs:
              total_energy_rmse: 1.0
              per_atom_energy_rmse: null
              forces_rmse: 1.0
              total_energy_mae: null
              per_atom_energy_mae: null
              forces_mae: null

    Args:
        coeffs (Dict[str, float]): ``dict`` that stores the relative contribution of the different energy and forces metrics to the ``weighted_sum`` version of the metric as in ``nequip.train.MetricsManager`` (default ``{'total_energy_rmse': 1.0, 'per_atom_energy_rmse': None, 'forces_rmse': 1.0, 'total_energy_mae': None, 'per_atom_energy_mae': None, 'forces_mae': None}``)
    """
    assert all(
        [k in _EF_METRICS_COEFFS_KEYS for k in coeffs.keys()]
    ), f"Unrecognized key found in `coeffs`, only the following are recognized: {_EF_METRICS_COEFFS_KEYS}"
    metrics = [
        {
            "name": "total_energy_rmse",
            "field": AtomicDataDict.TOTAL_ENERGY_KEY,
            "metric": RootMeanSquaredError(),
            "coeff": coeffs.get("total_energy_rmse", None),
        },
        {
            "name": "total_energy_mae",
            "field": AtomicDataDict.TOTAL_ENERGY_KEY,
            "metric": MeanAbsoluteError(),
            "coeff": coeffs.get("total_energy_mae", None),
        },
        {
            "name": "per_atom_energy_rmse",
            "field": PerAtomModifier(AtomicDataDict.TOTAL_ENERGY_KEY),
            "metric": RootMeanSquaredError(),
            "coeff": coeffs.get("per_atom_energy_rmse", None),
        },
        {
            "name": "per_atom_energy_mae",
            "field": PerAtomModifier(AtomicDataDict.TOTAL_ENERGY_KEY),
            "metric": MeanAbsoluteError(),
            "coeff": coeffs.get("per_atom_energy_mae", None),
        },
        {
            "name": "forces_rmse",
            "field": AtomicDataDict.FORCE_KEY,
            "metric": RootMeanSquaredError(),
            "coeff": coeffs.get("forces_rmse", None),
        },
        {
            "name": "forces_mae",
            "field": AtomicDataDict.FORCE_KEY,
            "metric": MeanAbsoluteError(),
            "coeff": coeffs.get("forces_mae", None),
        },
    ]
    return MetricsManager(metrics, type_names=type_names)


def EnergyForceStressLoss(
    coeffs: Dict[str, float] = {
        AtomicDataDict.TOTAL_ENERGY_KEY: 1.0,
        AtomicDataDict.FORCE_KEY: 1.0,
        AtomicDataDict.STRESS_KEY: 1.0,
    },
    per_atom_energy: bool = True,
    type_names=None,
):
    """Simplified :class:`MetricsManager` wrapper for a **loss** term containing energy, forces and stress mean squared errors (MSEs).

    The loss component names are ``per_atom_energy_mse`` OR ``total_energy_mse`` (depending on whether ``per_atom_energy`` is ``True`` or ``False``), ``forces_mse``, and ``stress_mse``, which are the names to refer to when neeeded, e.g. when scheduling loss component coefficients.

    Example usage in config:

    .. code-block:: yaml

        training_module:
          _target_: nequip.train.NequIPLightningModule

          loss:
            _target_: nequip.train.EnergyForceStressLoss
            per_atom_energy: true
            coeffs:
              total_energy: 1.0
              forces: 1.0
              stress: 1.0

    Args:
        coeffs (Dict[str, float]): ``dict`` that stores the relative weight of energy and forces to the overall loss (default ``{'total_energy': 1.0, 'forces': 1.0, 'stress': 1.0}``)
        per_atom_energy (bool, optional): whether to normalize the total energy by the number of atoms (default ``True``)
    """

    metrics = [
        {
            "name": "per_atom_energy_mse" if per_atom_energy else "total_energy_mse",
            "field": (
                PerAtomModifier(AtomicDataDict.TOTAL_ENERGY_KEY)
                if per_atom_energy
                else AtomicDataDict.TOTAL_ENERGY_KEY
            ),
            "coeff": coeffs[AtomicDataDict.TOTAL_ENERGY_KEY],
            "metric": MeanSquaredError(),
        },
        {
            "name": "forces_mse",
            "field": AtomicDataDict.FORCE_KEY,
            "coeff": coeffs[AtomicDataDict.FORCE_KEY],
            "metric": MeanSquaredError(),
        },
        {
            "name": "stress_mse",
            "field": AtomicDataDict.STRESS_KEY,
            "coeff": coeffs[AtomicDataDict.STRESS_KEY],
            "metric": MeanSquaredError(),
        },
    ]
    return MetricsManager(metrics, type_names=type_names)


def EnergyForceStressMetrics(
    coeffs: Dict[str, float] = {
        "total_energy_rmse": 1.0,
        "per_atom_energy_rmse": None,
        "forces_rmse": 1.0,
        "stress_rmse": 1.0,
        "total_energy_mae": None,
        "per_atom_energy_mae": None,
        "forces_mae": None,
        "stress_mae": None,
    },
    type_names=None,
):
    """Simplified :class:`MetricsManager` wrapper for a **metric** term containing energy, force and stress mean absolute errors (MAEs) and root mean squared errors (RMSEs).

    Example usage in config:

    .. code-block:: yaml

        training_module:
          _target_: nequip.train.NequIPLightningModule

          val_metrics:
            _target_: nequip.train.EnergyForceStressMetrics
            coeffs:
              total_energy_rmse: 1.0
              per_atom_energy_rmse: null
              forces_rmse: 1.0
              stress_rmse: 1.0
              total_energy_mae: null
              per_atom_energy_mae: null
              forces_mae: null
              stress_mae: null

    Args:
        coeffs (Dict[str, float]): ``dict`` that stores the relative contribution of the different energy and forces metrics to the ``weighted_sum`` version of the metric as in ``nequip.train.MetricsManager`` (default ``{'total_energy_rmse': 1.0, 'per_atom_energy_rmse': None, 'forces_rmse': 1.0, 'stress_rmse': 1.0, 'total_energy_mae': None, 'per_atom_energy_mae': None, 'forces_mae': None, 'stress_mae': None}``)
    """
    assert all(
        [k in _EFS_METRICS_COEFFS_KEYS for k in coeffs.keys()]
    ), f"Unrecognized key found in `coeffs`, only the following are recognized: {_EFS_METRICS_COEFFS_KEYS}"
    metrics = [
        {
            "name": "total_energy_rmse",
            "field": AtomicDataDict.TOTAL_ENERGY_KEY,
            "metric": RootMeanSquaredError(),
            "coeff": coeffs.get("total_energy_rmse", None),
        },
        {
            "name": "total_energy_mae",
            "field": AtomicDataDict.TOTAL_ENERGY_KEY,
            "metric": MeanAbsoluteError(),
            "coeff": coeffs.get("total_energy_mae", None),
        },
        {
            "name": "per_atom_energy_rmse",
            "field": PerAtomModifier(AtomicDataDict.TOTAL_ENERGY_KEY),
            "metric": RootMeanSquaredError(),
            "coeff": coeffs.get("per_atom_energy_rmse", None),
        },
        {
            "name": "per_atom_energy_mae",
            "field": PerAtomModifier(AtomicDataDict.TOTAL_ENERGY_KEY),
            "metric": MeanAbsoluteError(),
            "coeff": coeffs.get("per_atom_energy_mae", None),
        },
        {
            "name": "forces_rmse",
            "field": AtomicDataDict.FORCE_KEY,
            "metric": RootMeanSquaredError(),
            "coeff": coeffs.get("forces_rmse", None),
        },
        {
            "name": "forces_mae",
            "field": AtomicDataDict.FORCE_KEY,
            "metric": MeanAbsoluteError(),
            "coeff": coeffs.get("forces_mae", None),
        },
        {
            "name": "stress_rmse",
            "field": AtomicDataDict.STRESS_KEY,
            "metric": RootMeanSquaredError(),
            "coeff": coeffs.get("stress_rmse", None),
        },
        {
            "name": "stress_mae",
            "field": AtomicDataDict.STRESS_KEY,
            "metric": MeanAbsoluteError(),
            "coeff": coeffs.get("stress_mae", None),
        },
    ]
    return MetricsManager(metrics, type_names=type_names)


def EnergyOnlyLoss(
    per_atom_energy: bool = True,
    type_names=None,
):
    """Simplified :class:`MetricsManager` wrapper for a **loss** term containing only energy mean squared error (MSE).

    The loss component name is ``per_atom_energy_mse`` OR ``total_energy_mse`` (depending on whether ``per_atom_energy`` is ``True`` or ``False``).

    Example usage in config:

    .. code-block:: yaml

        training_module:
          _target_: nequip.train.NequIPLightningModule

          loss:
            _target_: nequip.train.EnergyOnlyLoss
            per_atom_energy: true

    Args:
        per_atom_energy (bool, optional): whether to normalize the total energy by the number of atoms (default ``True``)
    """

    metrics = [
        {
            "name": "per_atom_energy_mse" if per_atom_energy else "total_energy_mse",
            "field": (
                PerAtomModifier(AtomicDataDict.TOTAL_ENERGY_KEY)
                if per_atom_energy
                else AtomicDataDict.TOTAL_ENERGY_KEY
            ),
            "coeff": 1.0,  # single term, coefficient is always 1
            "metric": MeanSquaredError(),
        },
    ]
    return MetricsManager(metrics, type_names=type_names)


_ENERGY_ONLY_METRICS_COEFFS_KEYS: Final[List[str]] = [
    "total_energy_rmse",
    "per_atom_energy_rmse",
    "total_energy_mae",
    "per_atom_energy_mae",
]


def EnergyOnlyMetrics(
    coeffs: Dict[str, float] = {
        "total_energy_rmse": 1.0,
        "per_atom_energy_rmse": None,
        "total_energy_mae": None,
        "per_atom_energy_mae": None,
    },
    type_names=None,
):
    """Simplified :class:`MetricsManager` wrapper for a **metric** term containing only energy mean absolute errors (MAEs) and root mean squared errors (RMSEs).

    Example usage in config:

    .. code-block:: yaml

        training_module:
          _target_: nequip.train.NequIPLightningModule

          val_metrics:
            _target_: nequip.train.EnergyOnlyMetrics
            coeffs:
              total_energy_rmse: 1.0
              per_atom_energy_rmse: null
              total_energy_mae: null
              per_atom_energy_mae: null

    Args:
        coeffs (Dict[str, float]): ``dict`` that stores the relative contribution of the different energy metrics to the ``weighted_sum`` version of the metric as in ``nequip.train.MetricsManager`` (default ``{'total_energy_rmse': 1.0, 'per_atom_energy_rmse': None, 'total_energy_mae': None, 'per_atom_energy_mae': None}``)
    """
    assert all(
        [k in _ENERGY_ONLY_METRICS_COEFFS_KEYS for k in coeffs.keys()]
    ), f"Unrecognized key found in `coeffs`, only the following are recognized: {_ENERGY_ONLY_METRICS_COEFFS_KEYS}"
    metrics = [
        {
            "name": "total_energy_rmse",
            "field": AtomicDataDict.TOTAL_ENERGY_KEY,
            "metric": RootMeanSquaredError(),
            "coeff": coeffs.get("total_energy_rmse", None),
        },
        {
            "name": "total_energy_mae",
            "field": AtomicDataDict.TOTAL_ENERGY_KEY,
            "metric": MeanAbsoluteError(),
            "coeff": coeffs.get("total_energy_mae", None),
        },
        {
            "name": "per_atom_energy_rmse",
            "field": PerAtomModifier(AtomicDataDict.TOTAL_ENERGY_KEY),
            "metric": RootMeanSquaredError(),
            "coeff": coeffs.get("per_atom_energy_rmse", None),
        },
        {
            "name": "per_atom_energy_mae",
            "field": PerAtomModifier(AtomicDataDict.TOTAL_ENERGY_KEY),
            "metric": MeanAbsoluteError(),
            "coeff": coeffs.get("per_atom_energy_mae", None),
        },
    ]
    return MetricsManager(metrics, type_names=type_names)



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from typing import Union, Dict, List, Callable

import torch

from .. import AtomicDataDict
from ..dict import from_dict
from .base_datasets import AtomicDataset


class HDF5Dataset(AtomicDataset):
    """:class:`~nequip.data.dataset.AtomicDataset` that loads data from a HDF5 file.

    This class is useful for very large datasets that cannot fit in memory. It
    efficiently loads data from disk as needed without everything needing to be
    in memory at once.

    To use this, ``file_name`` should point to the HDF5 file, or alternatively a
    semicolon separated list of multiple files.  Each group in the file contains
    samples that all have the same number of atoms.  Typically there is one
    group for each unique number of atoms, but that is not required.  Each group
    should contain arrays whose length equals the number of samples, one for each
    type of data.  The names of the arrays can be specified with ``key_mapping``.

    Args:
        file_name (str): a semicolon separated list of HDF5 files.
        transforms (List[Callable]): list of data transforms
        key_mapping (Dict[str, str]): mapping of array names in the HDF5 file to ``AtomicDataDict`` keys
    """

    def __init__(
        self,
        file_name: str,
        transforms: List[Callable] = [],
        key_mapping: Dict[str, str] = {
            "pos": AtomicDataDict.POSITIONS_KEY,
            "energy": AtomicDataDict.TOTAL_ENERGY_KEY,
            "forces": AtomicDataDict.FORCE_KEY,
            "atomic_numbers": AtomicDataDict.ATOMIC_NUMBERS_KEY,
            "types": AtomicDataDict.ATOM_TYPE_KEY,
        },
    ):
        super().__init__(transforms=transforms)
        self.file_name = file_name
        self.key_mapping = key_mapping
        self.index = None
        self.num_frames = 0

        import h5py

        files = [h5py.File(f, "r") for f in self.file_name.split(";")]
        for file in files:
            for group_name in file:
                for key in self.key_mapping.keys():
                    if key in file[group_name]:
                        self.num_frames += len(file[group_name][key])
                        break
            file.close()

    def setup_index(self):
        import h5py

        files = [h5py.File(f, "r") for f in self.file_name.split(";")]
        self.has_forces = False
        self.index = []
        for file in files:
            for group_name in file:
                group = file[group_name]
                values = [None] * len(self.key_mapping.keys())
                samples = 0
                for i, key in enumerate(self.key_mapping.keys()):
                    if key in group:
                        values[i] = group[key]
                        samples = len(values[i])
                for i in range(samples):
                    self.index.append(tuple(values + [i]))

    def __len__(self) -> int:
        return self.num_frames

    def get_data_list(
        self,
        indices: Union[List[int], torch.Tensor, slice],
    ) -> List[AtomicDataDict.Type]:
        if self.index is None:
            self.setup_index()
        if isinstance(indices, slice):
            indices = range(len(self))[indices]
        return [self._get_data(index) for index in indices]

    def _get_data(self, idx: int) -> AtomicDataDict:
        data = self.index[idx]
        i = data[-1]
        data_dict = {}
        for j, value in enumerate(self.key_mapping.values()):
            if data[j] is not None:
                data_dict[value] = data[j][i]
        return from_dict(data_dict)



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import torch

from e3nn.o3._irreps import Irreps
from e3nn.nn._gate import Gate
from e3nn.nn._normact import NormActivation

from nequip.data import AtomicDataDict
from ._graph_mixin import GraphModuleMixin
from .interaction_block import InteractionBlock
from .nonlinearities import shifted_softplus
from .utils import tp_path_exists

from typing import Dict, Callable


acts = {
    "abs": torch.abs,
    "tanh": torch.tanh,
    "ssp": shifted_softplus,
    "silu": torch.nn.functional.silu,
}


class ConvNetLayer(GraphModuleMixin, torch.nn.Module):
    """
    Args:

    """

    resnet: bool

    def __init__(
        self,
        irreps_in,
        feature_irreps_hidden,
        convolution=InteractionBlock,
        convolution_kwargs: dict = {},
        resnet: bool = False,
        nonlinearity_type: str = "gate",
        nonlinearity_scalars: Dict[int, Callable] = {"e": "silu", "o": "tanh"},
        nonlinearity_gates: Dict[int, Callable] = {"e": "silu", "o": "tanh"},
    ):
        super().__init__()
        # initialization
        assert nonlinearity_type in ("gate", "norm")
        # make the nonlin dicts from parity ints instead of convinience strs
        nonlinearity_scalars = {
            1: nonlinearity_scalars["e"],
            -1: nonlinearity_scalars["o"],
        }
        nonlinearity_gates = {
            1: nonlinearity_gates["e"],
            -1: nonlinearity_gates["o"],
        }

        self.feature_irreps_hidden = Irreps(feature_irreps_hidden)
        self.resnet = resnet

        # We'll set irreps_out later when we know them
        self._init_irreps(
            irreps_in=irreps_in,
            required_irreps_in=[AtomicDataDict.NODE_FEATURES_KEY],
        )

        edge_attr_irreps = self.irreps_in[AtomicDataDict.EDGE_ATTRS_KEY]
        irreps_layer_out_prev = self.irreps_in[AtomicDataDict.NODE_FEATURES_KEY]

        irreps_scalars = Irreps(
            [
                (mul, ir)
                for mul, ir in self.feature_irreps_hidden
                if ir.l == 0
                and tp_path_exists(irreps_layer_out_prev, edge_attr_irreps, ir)
            ]
        )

        irreps_gated = Irreps(
            [
                (mul, ir)
                for mul, ir in self.feature_irreps_hidden
                if ir.l > 0
                and tp_path_exists(irreps_layer_out_prev, edge_attr_irreps, ir)
            ]
        )

        irreps_layer_out = (irreps_scalars + irreps_gated).simplify()

        if nonlinearity_type == "gate":
            ir = (
                "0e"
                if tp_path_exists(irreps_layer_out_prev, edge_attr_irreps, "0e")
                else "0o"
            )
            irreps_gates = Irreps([(mul, ir) for mul, _ in irreps_gated])

            # TO DO, it's not that safe to directly use the
            # dictionary
            equivariant_nonlin = Gate(
                irreps_scalars=irreps_scalars,
                act_scalars=[
                    acts[nonlinearity_scalars[ir.p]] for _, ir in irreps_scalars
                ],
                irreps_gates=irreps_gates,
                act_gates=[acts[nonlinearity_gates[ir.p]] for _, ir in irreps_gates],
                irreps_gated=irreps_gated,
            )

            conv_irreps_out = equivariant_nonlin.irreps_in.simplify()

        else:
            conv_irreps_out = irreps_layer_out.simplify()

            equivariant_nonlin = NormActivation(
                irreps_in=conv_irreps_out,
                # norm is an even scalar, so use nonlinearity_scalars[1]
                scalar_nonlinearity=acts[nonlinearity_scalars[1]],
                normalize=True,
                epsilon=1e-8,
                bias=False,
            )

        self.equivariant_nonlin = equivariant_nonlin

        # TODO: partial resnet?
        if irreps_layer_out == irreps_layer_out_prev and resnet:
            # We are doing resnet updates and can for this layer
            self.resnet = True
        else:
            self.resnet = False

        # TODO: last convolution should go to explicit irreps out

        # override defaults for irreps:
        convolution_kwargs.pop("irreps_in", None)
        convolution_kwargs.pop("irreps_out", None)
        self.conv = convolution(
            irreps_in=self.irreps_in,
            irreps_out=conv_irreps_out,
            **convolution_kwargs,
        )

        # The output features are whatever we got in
        # updated with whatever the convolution outputs (which is a full graph module)
        self.irreps_out.update(self.conv.irreps_out)
        # but with the features updated by the nonlinearity
        self.irreps_out[AtomicDataDict.NODE_FEATURES_KEY] = (
            self.equivariant_nonlin.irreps_out
        )

    def forward(self, data: AtomicDataDict.Type) -> AtomicDataDict.Type:
        # save old features for resnet
        old_x = data[AtomicDataDict.NODE_FEATURES_KEY]
        # run convolution
        data = self.conv(data)
        # do nonlinearity
        data[AtomicDataDict.NODE_FEATURES_KEY] = self.equivariant_nonlin(
            data[AtomicDataDict.NODE_FEATURES_KEY]
        )
        # do resnet
        if self.resnet:
            data[AtomicDataDict.NODE_FEATURES_KEY] = (
                old_x + data[AtomicDataDict.NODE_FEATURES_KEY]
            )
        return data



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import numpy as np
import torch

from .. import AtomicDataDict
from ..dict import from_dict
from .base_datasets import AtomicDataset

from typing import Union, Dict, List, Callable


class NPZDataset(AtomicDataset):
    """:class:`~nequip.data.dataset.AtomicDataset` that loads data from an NPZ file following `sGDML <https://www.sgdml.org/#datasets>`_ conventions.
    It is also compatible with other datasets such as rMD-17, with a change in ``key_mapping``
    (the default ``key_mapping`` is set to be compatible with sGDML datasets).

    The ``NPZDataset`` avoids loading the whole dataset into memory.

    Args:
        file_path (str): path to npz file
        transforms (List[Callable]): list of data transforms
        key_mapping (Dict[str, str]): mapping of array names in the npz file to ``AtomicDataDict`` keys
    """

    def __init__(
        self,
        file_path: str,
        transforms: List[Callable] = [],
        key_mapping: Dict[str, str] = {
            "R": AtomicDataDict.POSITIONS_KEY,
            "z": AtomicDataDict.ATOMIC_NUMBERS_KEY,
            "E": AtomicDataDict.TOTAL_ENERGY_KEY,
            "F": AtomicDataDict.FORCE_KEY,
        },
    ):
        super().__init__(transforms=transforms)
        self.file_path = file_path
        self.key_mapping = key_mapping

        # use energy array to get num_frames (small to load into memory)
        E_key = None
        for k, v in key_mapping.items():
            if v == AtomicDataDict.TOTAL_ENERGY_KEY:
                E_key = k
        assert (
            E_key is not None
        ), "No key corresponding to `total_energy` found in npz dataset"
        with np.load(self.file_path, mmap_mode="r") as npz_data:
            self.num_frames = npz_data[E_key].shape[0]

    def __len__(self) -> int:
        return self.num_frames

    def get_data_list(
        self,
        indices: Union[List[int], torch.Tensor, slice],
    ) -> List[AtomicDataDict.Type]:

        if isinstance(indices, slice):
            indices = list(range(*indices.indices(self.num_frames)))

        # memory-map the file
        with np.load(self.file_path, mmap_mode="r") as npz_data:
            data_list = []
            for idx in indices:
                data_dict = {}
                for k, v in self.key_mapping.items():
                    # special case for now, may generalize if needed in the future
                    if v == AtomicDataDict.ATOMIC_NUMBERS_KEY:
                        data_dict[v] = npz_data[k]
                    else:
                        data_dict[v] = npz_data[k][idx]
                data_list.append(from_dict(data_dict))
            return data_list



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.

import torch

from nequip.model.saved_models.checkpoint import data_dict_from_checkpoint
from nequip.model.saved_models.package import (
    _get_shared_importer,
    _get_package_metadata,
    _suppress_package_importer_warnings,
)
from nequip.model.saved_models import ModelFromCheckpoint
from nequip.model.utils import (
    _COMPILE_MODE_OPTIONS,
    _EAGER_MODEL_KEY,
)
from nequip.nn.model_modifier_utils import is_persistent_model_modifier
from nequip.model.modify_utils import get_all_modifiers, only_apply_persistent_modifiers
from nequip.utils.logger import RankedLogger
from nequip.utils.versions import get_current_code_versions, _TORCH_GE_2_6
from nequip.utils.versions.version_utils import get_version_safe
from nequip.utils.global_state import set_global_state

from ._workflow_utils import set_workflow_state
from ._package_utils import (
    _EXTERNAL_MODULES,
    _MOCK_MODULES,
    _INTERNAL_MODULES,
)

from omegaconf import OmegaConf
import hydra
import argparse
import pathlib
import yaml


# === setup logging ===
hydra.core.utils.configure_log(None)
logger = RankedLogger(__name__, rank_zero_only=True)


# `nequip-package` generates the archival format for NequIP framework models. This file contains the information necessary to track the archival format itself.
# whenever the archival format changes, `_CURRENT_NEQUIP_PACKAGE_VERSION` (counter to track the packaged model format) should be bumped up to the next number. We can then condition `ModelFromPackage` on the packaging format version to decide code paths to load the model appropriately.
# `nequip-package` format version index to condition other features upon when loading `nequip-package` from a specific version
#
# Package version high-level CHANGELOG:
# (use git blame on this line to identify specific commits and details of changes)
# 0:
#   - Initial version
# 1:
#   - package_metadata.txt instead of package_metadata.pkl
# 2:
#   - added `external_modules`
_CURRENT_NEQUIP_PACKAGE_VERSION = 2


def main(args=None):

    parser = argparse.ArgumentParser(description="Package NequIP ecosystem models.")

    subparsers = parser.add_subparsers(dest="command", title="commands")

    build_parser = subparsers.add_parser("build", help="build a packaged model file")
    build_parser.add_argument(
        "ckpt_path",
        help="path to checkpoint file",
        type=str,
    )
    build_parser.add_argument(
        "output_path",
        help="output path to save the packaged model. NOTE: a `.nequip.zip` extension is mandatory",
        type=pathlib.Path,
    )

    info_parser = subparsers.add_parser(
        "info", help="get information from a packaged model file"
    )
    # positional argument
    info_parser.add_argument(
        "pkg_path",
        help="path to package file",
        type=str,
    )
    info_parser.add_argument(
        "--get-modifiers",
        help="print all available model modifiers in the package file",
        action="store_true",
    )
    info_parser.add_argument(
        "--yaml",
        help="output in YAML format",
        action="store_true",
    )

    args = parser.parse_args(args=args)

    if args.command == "info":
        assert str(args.pkg_path).endswith(
            ".nequip.zip"
        ), "packed model file to inspect must end with the `.nequip.zip` extension"

        with _suppress_package_importer_warnings():
            imp = torch.package.PackageImporter(args.pkg_path)
            pkg_metadata = _get_package_metadata(imp)

            # Load and process modifiers if requested
            modifiers_info = None
            if args.get_modifiers:
                model = imp.load_pickle(
                    package="model",
                    resource=f"{_EAGER_MODEL_KEY}_model.pkl",
                    map_location="cpu",
                )
                modifiers_info = [
                    {
                        "name": name,
                        "persistent": is_persistent_model_modifier(modifier),
                        "doc": (
                            modifier.__doc__ if modifier.__doc__ is not None else ""
                        ).strip(),
                    }
                    for name, modifier in get_all_modifiers(model).items()
                ]

            # Print output
            if args.yaml:
                output_data = {"package_metadata": pkg_metadata}
                if modifiers_info is not None:
                    output_data["modifiers"] = modifiers_info
                print(yaml.dump(output_data))
            else:
                print("Package Metadata")
                print("================")
                print(
                    yaml.dump(
                        pkg_metadata,
                        default_flow_style=False,
                    )
                )
                if modifiers_info is not None:
                    print("Available Modifiers")
                    print("===================")
                    for idx, modifier_info in enumerate(modifiers_info):
                        persistent_flag = "" if modifier_info["persistent"] else "non-"
                        print(
                            f"{idx + 1}. {modifier_info['name']}\t({persistent_flag}persistent)\n"
                        )
                        if "doc" in modifier_info:
                            print(f"\t{modifier_info['doc']}\n")

        return

    elif args.command == "build":

        set_workflow_state("package")

        assert str(args.output_path).endswith(
            ".nequip.zip"
        ), "output path must end with the `.nequip.zip` extension"

        # === handle internal and external modules ===
        overlap = set(_INTERNAL_MODULES) & set(_EXTERNAL_MODULES)
        assert (
            not overlap
        ), f"Internal and external modules overlap with the following module(s): {overlap}"

        logger.debug("Internal Modules: " + str(_INTERNAL_MODULES))
        logger.debug("External Modules: " + str(_EXTERNAL_MODULES))
        logger.debug("Mock Modules: " + str(_MOCK_MODULES))

        # === load checkpoint and extract info ===
        checkpoint = torch.load(
            args.ckpt_path,
            map_location="cpu",
            weights_only=False,
        )

        # === text files to save ===
        # == original config ==
        orig_config = checkpoint["hyper_parameters"]["info_dict"].copy()
        # remove version info
        orig_config.pop("versions")
        orig_config = OmegaConf.to_yaml(orig_config)
        code_versions = get_current_code_versions()

        # == set global state ==
        set_global_state()

        # == get example data from checkpoint ==
        # the reason for including it here is that whoever receives the packaged model file does not need to have access to the original data source to do `nequip-compile` on the packaged model (AOT export requires example data)
        logger.info("Instantiating datamodule for packaging.")
        data = data_dict_from_checkpoint(args.ckpt_path)

        # === perform packaging ===

        # the main complication is that we need to account for the possibility that the checkpoint is based on another packaged model
        # we then need to include the importer in the exporter for repackaging
        # see https://pytorch.org/docs/stable/package.html#re-export-an-imported-object
        # the reason for the present solution is due to the following constraints
        # 1. we need every model to come from the same importer, which is used for constructing the exporter (we won't be able to save models coming from different importers, even if we pass all the importers during exporter construction)
        # 2. we need to use `ModelFromPackage` in a manner consistent with the API of fresh model building or `ModelFromCheckpoint` to ensure correct state restoration when loading the model from checkpoint (note that `nequip-package` is always loading a model from checkpoint with `ModelFromCheckpoint`)

        # the present solution is to first load an eager model from checkpoint,
        # which tells us whether the model originates from a package or not
        # if it's not from a package, we just instantiate all model build types and package
        # if it's from a package, we have logic in `ModelFromPackage` to reuse the same `Importer` as the one used to load the eager model [this ensures that all models come from the same `Importer`]
        # we use `ModelFromCheckpoint` to load the other models to make sure the relevant modifiers are applied and the states are restored correctly by the checkpoint load

        # == get eager model ==
        # if the origin is `ModelFromPackage`, this call would have populated a variable that we can query later
        logger.info(f"Building `{_EAGER_MODEL_KEY}` model for packaging ...")
        with only_apply_persistent_modifiers(persistent_only=True):
            eager_model = ModelFromCheckpoint(
                args.ckpt_path, compile_mode=_EAGER_MODEL_KEY
            )

        # it's a `ModuleDict`, so we just reach into one of the models to get `type_names`
        # we expect all models to have the same `type_names` (see init of base Lightning module in `nequip/train/lightning.py`)
        type_names = list(eager_model.values())[0].type_names

        # == load models ==
        importers = (torch.package.importer.sys_importer,)
        models_to_package = {_EAGER_MODEL_KEY: eager_model}

        # this variable is None if the origin is not a package
        imp = _get_shared_importer()
        if imp is not None:
            # the origin is `ModelFromPackage`
            # first update the `importers`
            importers = (imp,) + importers
            # we only repackage what's in the package
            # e.g. if the original package was made with torch<2.6, and we're doing the current packaging with torch>=2.6, we'll miss the `compile` model, but there's nothing we can do about it
            package_compile_modes = _get_package_metadata(imp)["available_models"]
        else:
            if _TORCH_GE_2_6:
                # allow everything (including compile models)
                package_compile_modes = _COMPILE_MODE_OPTIONS.copy()
            else:
                # only allow eager model if not torch>=2.6
                package_compile_modes = [_EAGER_MODEL_KEY]

        # remove eager model since we already built it
        package_compile_modes.remove(_EAGER_MODEL_KEY)

        for compile_mode in package_compile_modes:
            logger.info(f"Building `{compile_mode}` model for packaging ...")
            with only_apply_persistent_modifiers(persistent_only=True):
                model = ModelFromCheckpoint(args.ckpt_path, compile_mode=compile_mode)
            models_to_package.update({compile_mode: model})

        # == package ==
        with _suppress_package_importer_warnings():

            with torch.package.PackageExporter(
                args.output_path, importer=importers, debug=True
            ) as exp:
                # handle dependencies
                exp.mock([f"{pkg}.**" for pkg in _MOCK_MODULES])
                exp.extern([f"{pkg}.**" for pkg in _EXTERNAL_MODULES])
                exp.intern([f"{pkg}.**" for pkg in _INTERNAL_MODULES])
                # add data for aotinductor compilation
                exp.save_pickle(
                    package="model",
                    resource="example_data.pkl",
                    obj=data,
                    dependencies=True,
                )
                # save misc info
                exp.save_text("model", "config.yaml", orig_config)

                # save metadata used for loading packages
                pkg_metadata = {
                    "versions": code_versions,
                    "external_modules": {
                        k: get_version_safe(k) for k in _EXTERNAL_MODULES
                    },
                    "package_version_id": _CURRENT_NEQUIP_PACKAGE_VERSION,
                    "available_models": list(models_to_package.keys()),
                    "atom_types": {idx: name for idx, name in enumerate(type_names)},
                }
                pkg_metadata = yaml.dump(
                    pkg_metadata,
                    default_flow_style=False,
                )
                exp.save_text("model", "package_metadata.txt", pkg_metadata)

                # save models
                for compile_mode, model in models_to_package.items():
                    exp.save_pickle(
                        package="model",
                        resource=f"{compile_mode}_model.pkl",
                        obj=model,
                        dependencies=True,
                    )

                del importers

        logger.info(f"Packaged model saved to {args.output_path}")
        set_workflow_state(None)
        return


if __name__ == "__main__":
    main()



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import torch
import packaging.version

from .version_utils import get_version_safe


_TORCH_VERSION = packaging.version.parse(get_version_safe(torch.__name__))
_TORCH_GE_2_4 = packaging.version.parse(
    _TORCH_VERSION.base_version
) >= packaging.version.parse(
    "2.4"
)  # open equivariance's lowest version
_TORCH_GE_2_6 = packaging.version.parse(
    _TORCH_VERSION.base_version
) >= packaging.version.parse("2.6")
_TORCH_GE_2_8 = packaging.version.parse(
    _TORCH_VERSION.base_version
) >= packaging.version.parse("2.8")


def check_pt2_compile_compatibility():
    assert _TORCH_GE_2_6, (
        f"PyTorch >= 2.6 required for PT2 compilation functionality, "
        f"but {_TORCH_VERSION} found."
    )



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from math import sqrt, prod
import torch

from e3nn.o3._irreps import Irreps
from e3nn.util.jit import compile_mode

from nequip.data import AtomicDataDict
from ._graph_mixin import GraphModuleMixin
from .nonlinearities import ShiftedSoftplus

from typing import Optional


@compile_mode("script")
class ScalarMLP(GraphModuleMixin, torch.nn.Module):
    """Apply an MLP to some scalar field."""

    field: str
    out_field: str

    def __init__(
        self,
        output_dim: int,
        hidden_layers_depth: int = 0,
        hidden_layers_width: Optional[int] = None,
        nonlinearity: Optional[str] = "silu",
        bias: bool = False,
        forward_weight_init: bool = True,
        init_mode: str = "uniform",
        field: str = AtomicDataDict.NODE_FEATURES_KEY,
        out_field: Optional[str] = None,
        irreps_in=None,
    ):
        super().__init__()
        self.field = field
        self.out_field = out_field if out_field is not None else field
        self._init_irreps(
            irreps_in=irreps_in,
            required_irreps_in=[self.field],
        )

        assert len(self.irreps_in[self.field]) == 1
        assert self.irreps_in[self.field][0].ir == (0, 1)  # scalars
        self.mlp_module = ScalarMLPFunction(
            input_dim=self.irreps_in[self.field][0].mul,
            output_dim=output_dim,
            hidden_layers_depth=hidden_layers_depth,
            hidden_layers_width=hidden_layers_width,
            nonlinearity=nonlinearity,
            bias=bias,
            forward_weight_init=forward_weight_init,
            init_mode=init_mode,
        )
        self.irreps_out[self.out_field] = Irreps([(self.mlp_module.dims[-1], (0, 1))])

    def forward(self, data: AtomicDataDict.Type) -> AtomicDataDict.Type:
        data[self.out_field] = self.mlp_module(data[self.field])
        return data


@compile_mode("script")
class ScalarMLPFunction(torch.nn.Module):
    """Module implementing an MLP according to provided options.

    ``input_dim`` and ``output_dim`` are mandatory arguments.
    If only ``input_dim`` and ``output_dim`` are specified, this module defaults to a linear layer (corresponding to the default of ``hidden_layers_depth=0``).
    If ``hidden_layers_depth!=0``,  ``hidden_layers_width`` must be configured (an error will be raised if the default of ``hidden_layers_width=None`` is used).

    Args:
        nonlinearity (str): ``silu`` (default), ``mish``, ``gelu``, or ``None``
        bias (bool): whether a bias is included (default ``False``)
        forward_weight_init (bool): whether to initialize weights to preserve forward activation variance (default ``True``) or initialize weights to preserve backward gradient variance
    """

    num_layers: int
    bias: bool
    is_nonlinear: bool

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_layers_depth: int = 0,
        hidden_layers_width: Optional[int] = None,
        nonlinearity: Optional[str] = "silu",
        bias: bool = False,
        forward_weight_init: bool = True,
        init_mode: str = "uniform",
    ):
        super().__init__()
        self.bias = bias

        # === process MLP dims ===
        if hidden_layers_depth != 0:
            assert hidden_layers_depth > 0 and hidden_layers_width > 0
        hidden_layers_dims = hidden_layers_depth * [hidden_layers_width]
        self.dims = [input_dim] + hidden_layers_dims + [output_dim]
        self.num_layers = len(self.dims) - 1
        assert self.num_layers >= 1
        # NOTE: `input_dim` and `output_dim` are always mandatory, which default to at least a linear
        # a one-layer MLP is a linear layer

        # === handle nonlinearity ===
        # TODO: maybe adapt gain to be nonlinearity dependent
        nonlinearity_module = {
            None: torch.nn.Identity,
            "silu": torch.nn.SiLU,
            "mish": torch.nn.Mish,
            "gelu": torch.nn.GELU,
            "ssp": ShiftedSoftplus,
            "tanh": torch.nn.Tanh,
        }[nonlinearity]
        self.is_nonlinear = False  # updated below in loop

        # === build the MLP + weight init ===
        mlp = torch.nn.Sequential()
        for layer, (h_in, h_out) in enumerate(zip(self.dims, self.dims[1:])):

            # === weight initialization ===
            # normalize to preserve variance of forward activations or backward derivatives
            # we use "relu" gain (sqrt(2)) as a stand-in for the smooth nonlinearities we use, and only apply them if there is a nonlinearity
            # for forward (backward) norm, we don't include the gain for the first (last) layer
            # see https://pytorch.org/docs/stable/nn.init.html#torch.nn.init.kaiming_uniform_
            if forward_weight_init:
                norm_dim = h_in
                gain = 1.0 if nonlinearity is None or (layer == 0) else sqrt(2)
            else:
                norm_dim = h_out
                gain = (
                    1.0
                    if nonlinearity is None or (layer == self.num_layers - 1)
                    else sqrt(2)
                )
            # === instantiate `Linear` ===
            linear_layer = ScalarLinearLayer(
                in_features=h_in,
                out_features=h_out,
                alpha=gain / sqrt(norm_dim),
                bias=bias,
                init_mode=init_mode,
            )
            mlp.append(linear_layer)
            del gain, norm_dim

            # === add nonlinearity (if any) except for last layer ===
            if (layer != self.num_layers - 1) and (nonlinearity is not None):
                # only update `self.is_nonlinear` when a nonlinearity is applied
                mlp.append(nonlinearity_module())
                self.is_nonlinear = True

        # the following attribute is only used for the "deep linear" code path in `forward`
        # its definition is not conditioned on the "deep linear" codepath for TorchScript compatibility

        # use `multidot` based implementation for deep linear net (no nonlinearity, no bias, more than one layer)
        # otherwise use the `mlp` built in init
        if (not self.is_nonlinear) and (not self.bias) and (self.num_layers > 1):
            self.mlp = DeepLinearMLP(mlp)
            del mlp
        else:
            self.mlp = mlp

    def forward(self, x):
        return self.mlp(x)


class DeepLinearMLP(torch.nn.Module):

    def __init__(self, mlp) -> None:

        super().__init__()
        self.weights = torch.nn.ParameterList()
        alphas = []
        for this_idx, mlp_idx in enumerate(range(len(mlp))):
            new_weight = torch.clone(mlp[mlp_idx].weight)
            self.weights.append(new_weight)
            del new_weight
            alphas.append(mlp[mlp_idx].alpha)
        self.alpha = prod(alphas)
        del alphas

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        weight = torch.mul(
            torch.linalg.multi_dot([weight for weight in self.weights]), self.alpha
        )
        return torch.mm(input, weight)


class ScalarLinearLayer(torch.nn.Module):
    """Module implementing a linear layer with a scaling factor `alpha` applied to the weights."""

    in_features: int
    out_features: int

    def __init__(
        self,
        in_features: int,
        out_features: int,
        alpha: float = 1.0,
        bias: bool = False,
        init_mode: str = "uniform",
    ) -> None:
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha
        self.weight = torch.nn.Parameter(torch.empty((in_features, out_features)))
        # initialize weights based on init_mode
        if init_mode == "uniform":
            # initialize weights to uniform distribution with mean 0 variance 1
            torch.nn.init.uniform_(self.weight, -sqrt(3), sqrt(3))
        elif init_mode == "normal":
            # initialize weights to normal distribution with mean 0 std 1
            torch.nn.init.normal_(self.weight, mean=0.0, std=1.0)
        else:
            raise ValueError(
                f"Unknown init_mode: {init_mode}. Must be 'uniform' or 'normal'."
            )
        # initialize bias (if any) to zeros
        if bias:
            self.bias = torch.nn.Parameter(torch.zeros(out_features))
        else:
            self.register_parameter("bias", None)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        weight = self.weight * self.alpha
        if self.bias is None:
            return torch.mm(input, weight)
        else:
            return torch.addmm(self.bias, input, weight)

    def extra_repr(self) -> str:
        return f"in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}"



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import torch
import lightning
from lightning.pytorch.callbacks import Callback

import ase

from nequip.data import AtomicDataDict, to_ase
from nequip.data import (
    _register_field_prefix,
    register_fields,
    _NODE_FIELDS,
    _EDGE_FIELDS,
    _GRAPH_FIELDS,
)
from nequip.train import NequIPLightningModule

from typing import List, Dict, Union, Optional


class TestTimeXYZFileWriter(Callback):
    """Writes model outputs to an ``xyz`` file.

    Users must provide an ``out_file`` that does not contain an extension. The actual output file will take
    the form ``{out_file}_dataset{idx}.xyz`` where ``idx`` is the dataset index (would be ``0`` for a single
    test set but varies depending on number of test sets).

    To incorporate original dataset fields in the ``xyz`` file to simplify analysis, users may provide
    ``output_fields_from_original_dataset``. Such fields will have the prefix ``original_dataset_`` in the ``xyz`` file.

    To obtain correct chemical species information, users must provide ``chemical_species`` in an order consistent with
    the model's ``type_names``.

    Example usage in config to write predictions and original dataset ``total_energy`` and ``forces`` to an ``xyz`` file:

    .. code-block:: yaml

        callbacks:
          - _target_: nequip.train.callbacks.TestTimeXYZFileWriter
            out_file: ${hydra:runtime.output_dir}/test
            output_fields_from_original_dataset: [total_energy, forces]
            chemical_symbols: ${chemical_symbols}

    Args:
        out_file (str): path to output file (must NOT contain ``.xyz`` or ``.extxyz`` extension)
        output_fields_from_original_dataset (List[str]): values from the original dataset to save in the ``out_file``
        chemical_species (List[str]): chemical species in the same order as model's ``type_names``
    """

    def __init__(
        self,
        out_file: str,
        output_fields_from_original_dataset: Optional[List[str]] = [],
        chemical_symbols: Optional[List[str]] = None,
    ):
        assert not (out_file.endswith(".xyz") or out_file.endswith(".extxyz"))
        self.out_file = out_file
        assert all(
            [
                field in (_NODE_FIELDS | _EDGE_FIELDS | _GRAPH_FIELDS)
                for field in output_fields_from_original_dataset
            ]
        )

        # special case total_energy (nequip's convention) vs energy (ase's convention)
        self.output_fields_from_original_dataset = []
        for field in output_fields_from_original_dataset:
            if field == "total_energy":
                self.output_fields_from_original_dataset.append("energy")
                register_fields(graph_fields=["original_dataset_energy"])
            else:
                self.output_fields_from_original_dataset.append(field)
        _register_field_prefix("original_dataset_")

        self.extra_fields = [
            "original_dataset_" + field
            for field in self.output_fields_from_original_dataset
        ]
        self.chemical_symbols = chemical_symbols

    def on_test_batch_end(
        self,
        trainer: lightning.Trainer,
        pl_module: NequIPLightningModule,
        outputs: Dict[str, Union[torch.Tensor, AtomicDataDict.Type]],
        batch: AtomicDataDict.Type,
        batch_idx: int,
        dataloader_idx=0,
    ):
        """"""
        with torch.no_grad():
            output_out = outputs[f"test_{dataloader_idx}_output"].copy()
            for field in self.output_fields_from_original_dataset:
                # special case total_energy (nequip's convention) vs energy (ase's convention)
                if field == "energy":
                    output_out["original_dataset_energy"] = batch["total_energy"]
                else:
                    output_out["original_dataset_" + field] = batch[field]

            # !! EXTREMELY IMPORTANT -- special handling of PBC key if present !!
            # ASE data inputs would possess it to be used at data precprocessing time (i.e. neighborlist construction)
            # but it won't be passed through the model, so we get it from `batch`
            if AtomicDataDict.PBC_KEY in batch:
                output_out[AtomicDataDict.PBC_KEY] = batch[AtomicDataDict.PBC_KEY]

            # append to the file
            ase.io.write(
                self.out_file + f"_dataset{dataloader_idx}.xyz",
                to_ase(
                    output_out,
                    chemical_symbols=self.chemical_symbols,
                    extra_fields=self.extra_fields,
                ),
                format="extxyz",
                append=True,
            )
            del output_out



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.

from .torchscript import load_torchscript_model
from .aotinductor import load_aotinductor_model
from .compiled import load_compiled_model

__all__ = ["load_torchscript_model", "load_aotinductor_model", "load_compiled_model"]



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import torch
import os

from nequip.nn.compile import ListInputOutputWrapper, DictInputOutputWrapper
from nequip.data import AtomicDataDict
from nequip.data._key_registry import get_dynamic_shapes
from .fx import nequip_make_fx
from .compile import prepare_model_for_compile
from .versions import check_pt2_compile_compatibility
from .dtype import test_model_output_similarity_by_dtype, _pt2_compile_error_message

from typing import List, Dict, Union, Any


def aot_export_model(
    model: torch.nn.Module,
    device: Union[str, torch.device],
    input_fields: List[str],
    output_fields: List[str],
    data: AtomicDataDict.Type,
    batch_map: Dict[str, torch.export.dynamic_shapes.Dim],
    output_path: str,
    inductor_configs: Dict[str, Any] = {},
    seed: int = 1,
) -> str:
    # === torch version check ===
    check_pt2_compile_compatibility()

    # defensively refresh the cache
    torch._dynamo.reset()

    # === preprocess model and make_fx ===
    model_to_trace = ListInputOutputWrapper(model, input_fields, output_fields)
    model_to_trace = prepare_model_for_compile(model_to_trace, device)

    fx_model = nequip_make_fx(
        model=model_to_trace,
        data={k: data[k] for k in input_fields},
        fields=input_fields,
        seed=seed,
    )

    # === perform export ===
    # == define dynamics dims ==
    dynamic_shapes = get_dynamic_shapes(input_fields, batch_map)

    # == export ==
    exported = torch.export.export(
        fx_model,
        (*[data[k] for k in input_fields],),
        dynamic_shapes=dynamic_shapes,
    )
    # NOTE: the following requires PyTorch 2.6
    out_path = torch._inductor.aoti_compile_and_package(
        exported,
        package_path=output_path,
        inductor_configs=inductor_configs,
    )
    assert out_path == output_path

    # === sanity check ===
    if os.environ.get("NEQUIP_SKIP_AOTI_MODEL_CHECK", "0") != "1":
        aot_model = DictInputOutputWrapper(
            torch._inductor.aoti_load_package(out_path),
            input_fields,
            output_fields,
        )
        test_model_output_similarity_by_dtype(
            aot_model,
            model,
            {k: data[k] for k in input_fields},
            model.model_dtype,
            fields=output_fields,
            error_message=_pt2_compile_error_message,
        )
        del aot_model

    return out_path



# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
from ._base_datamodule import NequIPDataModule
from nequip.data import AtomicDataDict
from nequip.utils import download_url, extract_zip
from nequip.utils.logger import RankedLogger

import os
from typing import Union, Sequence, List, Callable

logger = RankedLogger(__name__, rank_zero_only=True)

KCALMOL_TO_EV = 0.0433641


def _kcalmol_to_ev(data: AtomicDataDict.Type) -> AtomicDataDict.Type:
    data[AtomicDataDict.TOTAL_ENERGY_KEY] = (
        data[AtomicDataDict.TOTAL_ENERGY_KEY] * KCALMOL_TO_EV
    )
    data[AtomicDataDict.FORCE_KEY] = data[AtomicDataDict.FORCE_KEY] * KCALMOL_TO_EV
    return data


class MD22DataModule(NequIPDataModule):
    """Lightning Data Module responsible for processing sGDML MD22 datasets (including downloading).

    This class handles the MD22 datasets, including ``tetrapeptide`` (CHNO), ``dha`` (CHO), ``stachyose`` (CHO), ``dna_atat`` (CHNO), ``dna_atat_cgcg`` (CHNO), ``buckyball_catcher`` (CH), and ``double_walled_nanotube`` (CH).
    See `Science Advances 9.2 (2023): eadf0873 <https://www.science.org/doi/10.1126/sciadv.adf0873>`_ for more details.

    This datamodule will automatically use the training set sizes from the paper, that is, ``tetrapeptide`` (6,000/85,109), ``dha`` (8,000/69,753), ``stachyose`` (8,000/27,272), ``dna_atat`` (3,000/20,001), ``dna_atat_cgcg`` (2,000/10,153), ``buckyball_catcher`` (600/6,102), and ``double_walled_nanotube`` (800/5,032). The "training set" will then be partitioned into train and validation datasets based on ``train_val_split``. The remainder is used as the test dataset.

    Args:
        dataset (str): ``tetrapeptide``, ``dha``, ``stachyose``, ``dna_atat``, ``dna_atat_cgcg``, ``buckyball_catcher``, or ``double_walled_nanotube``
        data_source_dir (str): directory to download sGDML MD22 data to, or where the npz files are present if already downloaded
        transforms (List[Callable]): list of data transforms
        seed (int): data seed for reproducibility
        train_val_split (List[float] or List[int]): train-validation split either in fractions ``[1, 1-f]`` or integers ``[N_train, N_val]``
    """

    # dataset: [file_name, num_trainval, num_data]
    dataset_map = {
        "tetrapeptide": ["md22_Ac-Ala3-NHMe.npz", 6000, 85109],
        "dha": ["md22_DHA.npz", 8000, 69753],
        "stachyose": ["md22_stachyose.npz", 8000, 27272],
        "dna_atat": ["md22_AT-AT.npz", 3000, 20001],
        "dna_atat_cgcg": ["md22_AT-AT-CG-CG.npz", 2000, 10153],
        "buckyball_catcher": ["md22_buckyball-catcher.npz", 600, 6102],
        "double_walled_nanotube": ["md22_double-walled_nanotube.npz", 800, 5032],
    }
    url_dict = {
        "tetrapeptide": "http://www.quantum-machine.org/gdml/repo/datasets/md22_Ac-Ala3-NHMe.npz",
        "dha": "http://www.quantum-machine.org/gdml/repo/datasets/md22_DHA.npz",
        "stachyose": "http://www.quantum-machine.org/gdml/repo/datasets/md22_stachyose.npz",
        "dna_atat": "http://www.quantum-machine.org/gdml/repo/datasets/md22_AT-AT.npz",
        "dna_atat_cgcg": "http://www.quantum-machine.org/gdml/repo/datasets/md22_AT-AT-CG-CG.npz",
        "buckyball_catcher": "http://www.quantum-machine.org/gdml/repo/datasets/md22_buckyball-catcher.npz",
        "double_walled_nanotube": "http://www.quantum-machine.org/gdml/repo/datasets/md22_double-walled_nanotube.npz",
    }

    def __init__(
        self,
        dataset: str,
        data_source_dir: str,
        transforms: List[Callable],
        seed: int,
        train_val_split: Sequence[Union[int, float]],
        **kwargs,
    ):

        assert (
            dataset in self.dataset_map.keys()
        ), f"`dataset={dataset}` not supported, `dataset` can be any of {list(self.dataset_map.keys())}"

        data_file_path = "/".join([data_source_dir, self.dataset_map[dataset][0]])
        base_config = {
            "_target_": "nequip.data.dataset.NPZDataset",
            "file_path": data_file_path,
            "transforms": list(transforms) + [_kcalmol_to_ev],
            "key_mapping": {
                "R": AtomicDataDict.POSITIONS_KEY,
                "z": AtomicDataDict.ATOMIC_NUMBERS_KEY,
                "E": AtomicDataDict.TOTAL_ENERGY_KEY,
                "F": AtomicDataDict.FORCE_KEY,
            },
        }

        train_end_idx = self.dataset_map[dataset][1]
        test_length = self.dataset_map[dataset][2] - train_end_idx

        train_config = {
            "_target_": "nequip.data.dataset.SubsetByRandomSlice",
            "dataset": base_config,
            "start": 0,
            "length": train_end_idx,
            "seed": seed,
        }

        test_config = {
            "_target_": "nequip.data.dataset.SubsetByRandomSlice",
            "dataset": base_config,
            "start": train_end_idx,
            "length": test_length,
            "seed": seed,
        }

        super().__init__(
            seed=seed,
            split_dataset={
                "dataset": train_config,
                "train": train_val_split[0],
                "val": train_val_split[1],
            },
            test_dataset=test_config,
            **kwargs,
        )
        self.dataset = dataset
        self.data_source_dir = data_source_dir
        self.data_file_path = data_file_path

    def prepare_data(self):
        """"""
        if not os.path.isfile(self.data_file_path):
            # download and unzip
            download_path = download_url(
                self.url_dict[self.dataset], self.data_source_dir
            )
            if download_path.endswith(".zip"):
                extract_zip(download_path, self.data_source_dir)
        else:
            logger.info(f"Using existing data file `{self.data_file_path}`")



# This file is a part of the `nequip` package. Please see LICENSE and README
# at the root for information on using it.
import torch
from e3nn.util.jit import script

from nequip.nn import graph_model
from nequip.utils.compile import prepare_model_for_compile
from nequip.utils.global_state import TF32_KEY

from typing import Union, Tuple


def load_torchscript_model(
    compile_path: str,
    device: Union[str, torch.device] = "cpu",
) -> Tuple[torch.nn.Module, dict]:
    """Load a torchscript model from a .nequip.pth file.

    Args:
        compile_path (str): path to compiled model file ending with .nequip.pth
        device (Union[str, torch.device]): the device to use
    """
    # load model with metadata
    metadata = {
        graph_model.R_MAX_KEY: None,
        graph_model.TYPE_NAMES_KEY: None,
        graph_model.PER_EDGE_TYPE_CUTOFF_KEY: None,
        TF32_KEY: None,
    }
    model = torch.jit.load(compile_path, _extra_files=metadata, map_location=device)
    model = torch.jit.freeze(model)

    # process metadata
    metadata[graph_model.R_MAX_KEY] = float(metadata[graph_model.R_MAX_KEY])
    metadata[graph_model.TYPE_NAMES_KEY] = (
        metadata[graph_model.TYPE_NAMES_KEY].decode("utf-8").split(" ")
    )

    # process per-edge-type cutoffs if present
    if metadata[graph_model.PER_EDGE_TYPE_CUTOFF_KEY] is not None:
        from nequip.nn.embedding.utils import parse_per_edge_type_cutoff_metadata

        cutoff_str = metadata[graph_model.PER_EDGE_TYPE_CUTOFF_KEY].decode("utf-8")
        metadata[graph_model.PER_EDGE_TYPE_CUTOFF_KEY] = (
            parse_per_edge_type_cutoff_metadata(
                cutoff_str, metadata[graph_model.TYPE_NAMES_KEY]
            )
        )

    return model, metadata


def save_torchscript_model(
    model: torch.nn.Module,
    metadata: dict,
    output_path: str,
    device: Union[str, torch.device],
) -> None:
    """Save a model as a torchscript .nequip.pth file.

    Args:
        model: model to save
        metadata: metadata dictionary to save with the model
        output_path: path to save the compiled model
        device: device to prepare model on
    """
    # encode metadata for torchscript
    encoded_metadata = {k: str(v).encode("ascii") for k, v in metadata.items()}

    # prepare and script model
    model = prepare_model_for_compile(model, device)
    script_model = script(model)

    # save with metadata
    torch.jit.save(script_model, output_path, _extra_files=encoded_metadata)
